__all__ = ['rgb2hcm']

from skimage.transform import resize
from scipy.optimize import minimize
import numpy as np

_k = np.ones(3)


def rgb2hcm(image):
    '''\
    rgb2hcm(image)

    Segmentation of an object with homogeneous background.

    Parameters
    ----------
    image: 3 dimensional ndarray
        The RGB input image.

    Returns
    -------
    hcm: 2 dimensional ndarray
        high contrast grayscale representation of input image


    Examples
    --------
    (TODO)
    '''
    if image.ndim < 3:
        I = image
    else:
        img_resize = resize(image, (64, 64), order=3,
                            mode='reflect', anti_aliasing=False)
        k = minimize(monochrome_std, [1., 1.], args=(img_resize,))['x']
        _k[:2] = k
        I = image @ _k
    J = I - I.min()
    J = J / J.max()
    n = J.shape[0] // 4
    m = J.shape[1] // 4

    if (J[:n, :m].mean() > .4):
        J = 1 - J
    return J


def monochrome_std(k, image):
    _k[:2] = k
    I = image @ _k
    return - I.std(ddof=1) / (I.max() - I.min())



__all__ = ['clean']

import numpy as np
import warnings

def clean(features, show=False):
    n_features = features.shape[1]
    ip = np.ones(n_features, dtype=int)

    # cleaning correlated features
    warnings.filterwarnings('ignore')
    C = np.abs(np.corrcoef(features, rowvar=False))
    idxs = np.vstack(np.where(C > .99))
    
    # remove pairs of same feature ( feature i will have a correlation of 1 whit itself )
    idxs = idxs[:, idxs[0,:] != idxs[1,:]]
    
    # remove correlated features
    if idxs.size > 0:
        ip[np.max(idxs, 0)] = 0
    
    # remove constant features
    s = features.std(axis=0, ddof=1)
    ip[s < 1e-8] = 0
    p = np.where(ip.astype(bool))[0]

    if show:
        print(f'Clean: number of features reduced from {n_features} to {p.size}.')

    return p


__all__ = ['fst_deriv']

import numpy as np 
from scipy import signal

def fst_deriv(image, mask):
    if mask % 2 == 0:
        raise Exception('Mask must be of odd size')
    sigma = mask / 8.5
    c = (mask - 1) / 2
    x = np.tile(np.arange(mask) + 1 - c, (mask, 1))
    y = x.T
    exp = np.exp(-(x**2 + y**2) / 2 / sigma ** 2)
    gauss_x = x * exp
    gauss_y = y * exp
    mgx = np.abs(gauss_x).sum() / 2 * (0.3192 * mask - .3543)
    gauss_x /= mgx
    gauss_y /= mgx
    dx = signal.convolve2d(image, gauss_x, 'same')
    dy = signal.convolve2d(image, gauss_y, 'same')
    return np.sqrt(dx**2 + dy**2), dx, dy


__all__ = ['sp100']


def sp100(features, classification):
    norm = (classification.flatten() -  classification.min()).astype(bool)
    max_class = norm.max()

    if max_class > 1:
        raise Exception('sp100 works only for two classes')

    c1_features = features[norm == 1,:]

    min_feat = c1_features.min(0)
    max_feat = c1_features.max(0)

    z1 = (features >= min_feat) & (features <= max_feat)
    dr = z1.all(1)
    
    TP = (dr * norm).sum()
    FP = dr.sum() - TP

    TN = (~dr * ~norm).sum()
    return TN / (FP+TN)


from sklearn.base import BaseEstimator, TransformerMixin
import tqdm
from sklearn.pipeline import make_pipeline

__all__ = ['FeatureExtractor', 'ImageProcessor']


class FeatureExtractor(BaseEstimator, TransformerMixin):
    """Base class for all feature extractors in PyBalu


    Notes (from sklearn.base.BaseEstimator)
    -----
    All estimators should specify all the parameters that can be set
    at the class level in their ``__init__`` as explicit keyword
    arguments (no ``*args`` or ``**kwargs``).
    """

    def __init__(self):
        self.show = False

    def _get_iterator(self, X, *, desc=None):
        if self.show:
            if desc is not None:
                desc = f'Extracting Features({desc})'
            else:
                desc = 'Extracting Features'
            return tqdm.tqdm(X, desc=desc, total=len(X), unit_scale=True)
        return iter(X)

    def fit(self, X, y=None):
        """Do nothing and return the estimator unchanged
        This method is just there to implement the usual API and hence
        work in pipelines.
        Parameters
        ----------
        X : array-like, shape [n_samples, n_features]
            Training data.
        """
        return self


class ImageProcessor(BaseEstimator, TransformerMixin):
    """Base class for all image transformers in PyBalu


    Notes (from sklearn.base.BaseEstimator)
    -----
    All estimators should specify all the parameters that can be set
    at the class level in their ``__init__`` as explicit keyword
    arguments (no ``*args`` or ``**kwargs``).
    """

    def __init__(self):
        self.show = False

    def _get_iterator(self, X, *, desc=None):
        if self.show:
            if desc is not None:
                desc = f'Processing Images({desc})'
            else:
                desc = 'Processing Images'
            return tqdm.tqdm(X, desc=desc, total=len(X), unit_scale=True)
        return iter(X)

    def fit(self, X, y=None):
        """Do nothing and return the estimator unchanged
        This method is just there to implement the usual API and hence
        work in pipelines.
        Parameters
        ----------
        X : array-like, shape [n_samples, n_features]
            Training data.
        """
        return self



__all__ = ['snd_deriv']

import numpy as np

def snd_deriv(image):
    m = image.shape[0]
    n = image.shape[1]
    
    out = np.zeros((m,n))
    
    out = -4 * image
    out[1:, :] += image[:m-1, :]
    out[:m-1, :] += image[1:, :]
    out[:, 1:] += image[:, :n-1]
    out[:, :n-1] += image[:, 1:]
    return out


__all__ = ['hog_features', 'HOGExtractor']

import numpy as np
from scipy.signal import convolve2d
from pybalu.misc import im2row

from .hog_utils import sep_into_bins
from pybalu.base import FeatureExtractor

_hj = np.array([[1, 0, -1]])
_hi = - _hj.T


def hog_features(image, region=None, *, v_windows=0, h_windows=0, n_bins=0, normalize=False, labels=False, show=False):
    '''\
hog_features(image, region=None, *, v_windows=0, h_windows=0, n_bins=0, normalize=False, labels=False, show=False)

(TODO)

Parameters
----------
(TODO)

Returns
-------
(TODO)

See Also
--------
(TODO)

Examples
--------
(TODO)
'''
    if not isinstance(v_windows, int) or v_windows <= 0:
        raise ValueError(
            f"`v_windows` must be an integer value greater than 0")
    if not isinstance(h_windows, int) or h_windows <= 0:
        raise ValueError(
            f"`h_windows` must be an integer value greater than 0")
    if not isinstance(n_bins, int) or n_bins <= 0:
        raise ValueError(f"`n_bins` must be an integer value greater than 0")

    if region is None:
        region = np.ones_like(image)

    Gj = convolve2d(image, _hj, mode='same')
    Gi = convolve2d(image, _hi, mode='same')

    A = np.mod(np.arctan2(Gi, Gj), np.pi)
    G = np.sqrt((Gi * Gi + Gj * Gj))
    B = np.floor_divide(A * n_bins, np.pi).astype(int)
    mags = im2row(G, v_windows, h_windows)
    bins = im2row(B, v_windows, h_windows)

    features = []

    for mag, _bins in zip(mags, bins):
        curr_feats = sep_into_bins(mag, _bins, n_bins)
        norm = (curr_feats @ curr_feats) ** .5
        curr_feats /= norm or 1
        features.append(curr_feats)

    hog_feats = np.hstack(features)

    if normalize:
        hog_feats /= hog_feats.sum()

    if labels:
        return np.array(["HOG"] * hog_feats.size), hog_feats

    return hog_feats


class HOGExtractor(FeatureExtractor):
    def __init__(self, *, v_windows=0, h_windows=0, n_bins=0, normalize=False):
        self.v_windows = v_windows
        self.h_windows = h_windows
        self.n_bins = n_bins
        self.normalize = normalize

    def transform(self, X):
        params = self.get_params()
        return np.array([hog_features(x, **params) for x in self._get_iterator(X)])

    def get_labels(self):
        return "HOG"



__all__ = ['from_image_set']

from pybalu.base import FeatureExtractor
from sklearn.base import TransformerMixin, clone
from sklearn.pipeline import Pipeline
import numpy as np
import tqdm


def from_image_set(image_set, extractor, show=False):
    '''\
from_image_set(image_set, extractor, show=False)

Returns a 2-dimensional array on which each row represents a different image from the given
image set, and each column a feature extracted by one of the given extractors. 

Parameters
----------
image_set : an ImageSet object
    Represents the set of images over which to extract the selected features. All images
    within the set will have their features extracted.
extractor : an object that implements the sklearn transformer interface
    Can be any of the extractors and image processors defined by PyBalu. Other useful classes inclue
    `sklearn.pipeline.Pipeline` and `sklearn.pipeline.FeatureUnion` to combine multiple extractors.
show : bool, optional
    If set to true, a progress bar that shows the current and estimated completion time will 
    be shown during execution. Default is False.

Returns
-------
features : a 2-dimensional ndarray
    An array with len(`image_set`) rows and a number of columns that depend on the number and
    nature of the given FeatureExtractors
labels : a one dimensional string ndarray
    With the same length as number of columns in `features`, labels each column corresponding
    to the feature it represents.

Examples
--------
Extract the basic intensity features and haralick features for distance 3 for all the images
in a given set

>>> from pybalu.io import ImageSet, print_features
>>> from pybalu.feature_extraction import BasicIntExtractor, HaralickExtractor, from_image_set
>>> from sklearn.pipeline import make_union
>>> 
>>> image_set = ImageSet('./images/', extension='.png')
>>> basic_int, haralick = BasicIntExtractor(), HaralickExtractor(distance=3)
>>> extractors = make_union(basic_int, haralick)
>>> features = from_image_set(image_set, 
...                           extractors,
...                           show=True)
Extracting Features(basic_int): 100%|███████████████████████| 1.68k/1.68k [00:12<00:00, 138.2 imgs/s]
Extracting Features(haralick): 100%|███████████████████████| 1.68k/1.68k [00:13<00:00, 129.9 imgs/s]
>>> labels = np.hstack([basic_int.get_labels(), haralick.get_labels()])
>>> print_features(labels, features[0])
Intensity Mean        :  104.13101
Intensity StdDev      :  22.29537
Intensity Kurtosis    :  4.27058
Intensity Skewness    : -0.89083
Mean Laplacian        : -3.79495
Mean Boundary Gradient: -1.00000
Tx 1 , d 3  (mean)    :  0.27961
Tx 2 , d 3  (mean)    :  0.35839
Tx 3 , d 3  (mean)    :  330.11359
Tx 4 , d 3  (mean)    :  0.35839
Tx 5 , d 3  (mean)    :  0.86984
Tx 6 , d 3  (mean)    :  7.50346
Tx 7 , d 3  (mean)    :  5.94866
Tx 8 , d 3  (mean)    :  1.55481
Tx 9 , d 3  (mean)    :  1.80165
Tx 10, d 3  (mean)    :  0.07122
Tx 11, d 3  (mean)    :  0.64535
Tx 12, d 3  (mean)    : -0.37320
Tx 13, d 3  (mean)    :  0.55530
Tx 14, d 3  (mean)    :  1.44741
Tx 1 , d 3  (range)   :  0.31881
Tx 2 , d 3  (range)   :  0.57074
Tx 3 , d 3  (range)   :  336.75712
Tx 4 , d 3  (range)   :  0.57074
Tx 5 , d 3  (range)   :  0.91066
Tx 6 , d 3  (range)   :  7.54292
Tx 7 , d 3  (range)   :  6.05999
Tx 8 , d 3  (range)   :  1.63802
Tx 9 , d 3  (range)   :  1.97060
Tx 10, d 3  (range)   :  0.08353
Tx 11, d 3  (range)   :  0.79496
Tx 12, d 3  (range)   :  0.50197
Tx 13, d 3  (range)   :  0.66518
Tx 14, d 3  (range)   :  1.48388
    '''
    if not isinstance(extractor, (TransformerMixin, Pipeline)):
        raise TypeError(
            'extractor object must implement sklearn transformer interface'
        )

    old_params = extractor.get_params()
    new_params = dict((k, show) if k.endswith("show") else (k, v)
                      for k, v in old_params.items())
    extractor.set_params(new_params)
    features = extractor.transform(image_set)
    extractor.set_params(old_params)
    return features



from .all import all_features, AllExtractor
from .basic_geo import basic_geo_features, BasicGeoExtractor
from .basic_int import basic_int_features, BasicIntExtractor
from .flusser import flusser_features, FlusserExtractor
from .fourier import fourier_features, FourierExtractor
from .fourier_des import fourier_des_features, FourierDesExtractor
from .gabor import gabor_features, GaborExtractor
from .gupta import gupta_features, GuptaExtractor
from .haralick import haralick_features, HaralickExtractor
from .hog import hog_features, HOGExtractor
from .hugeo import hugeo_features, HuGeoExtractor
from .huint import huint_features, HuIntExtractor
from .lbp import lbp_features, LBPExtractor


from .from_image_set import from_image_set

__all__ = ["all_features", "AllExtractor",
           "basic_geo_features", "BasicGeoExtractor",
           "basic_int_features", "BasicIntExtractor",
           "flusser_features", "FlusserExtractor",
           "fourier_features", "FourierExtractor",
           "fourier_des_features", "FourierDesExtractor",
           "gabor_features", "GaborExtractor",
           "gupta_features", "GuptaExtractor",
           "haralick_features", "HaralickExtractor",
           "hog_features", "HOGExtractor",
           "hugeo_features", "HuGeoExtractor",
           "huint_features", "HuIntExtractor",
           "lbp_features", "LBPExtractor",
           "from_image_set"]



import numpy as np
import scipy.ndimage as img
# pylint: disable=no-name-in-module
from .geometric_utils import bbox, convex_area, moments, perimeter
# pylint: enable=no-name-in-module
from pybalu.base import FeatureExtractor

__all__ = ['basic_geo_features', 'BasicGeoExtractor']

geo_labels = ['center of grav i [px]',
              'center of grav j [px]',
              'Height [px]',
              'Width [px]',
              'Area [px]',
              'Perimeter [px]',
              'Roundness',
              'Danielsson factor',
              'Euler Number',
              'Equivalent Diameter [px]',
              'MajorAxisLength [px]',
              'MinorAxisLength [px]',
              'Orientation [deg]',
              'Solidity',
              'Extent',
              'Eccentricity',
              'Convex Area [px]',
              'Filled Area [px]']


def basic_geo_features(image, *, show=False, labels=False):
    '''\
basic_geo_features(image, *, show=False, labels=False)

Return an array of with the standard geometric features of a binary image.

Parameters
----------
image : a numpy 2 dimensional int array
    It represents a binary image. Non binary arrays will return nonsensical results.
show : bool, optional
    Wether to print or not messages during execution
labels : bool, optional
    Wether to return a second array that contains the label of each value. 

Returns
-------
labels: ndarray, optional
    A one dimensional string ndarray that contains the labels to each of the features.
    This array is only returned if `labels` is set to True.
features: ndarray
    A float ndarray that contains all 18 geometric features extracted from `image`.

See Also
--------
fit_ellipse : <Not Implemented Yet>
flusser : Extract the four Flusser moments from binary image.
gupta : Extract the three Gupta moments from binary image.
hugeo : Extract the seven Hu moments from a binary image.

Examples
--------
Load an image and get its binary representation, then proceed to get its features:

>>> from pybalu.feature_extraction import basic_geo_features
>>> from pybalu.img_processing import segbalu
>>> from pybalu.io import imread
>>> img = imread('testimg.png')
>>> binary_img, _, _ = segbalu(img)
>>> features = basic_geo_features(binary_img)

Print a binary image features:

>>> from pybalu.io import print_features
>>> labels, features = basic_geo_features(binary_img, labels=True)
>>> print_features(labels, features)
center of grav i [px]   :  51.49202
center of grav j [px]   :  74.95331
Height [px]             :  72.00000
Width [px]              :  90.00000
Area [px]               :  2506.00000
Perimeter [px]          :  534.00000
Roundness               :  0.11044
Danielsson factor       :  13.58911
Euler Number            :  1.00000
Equivalent Diameter [px]:  56.48662
MajorAxisLength [px]    :  97.43620
MinorAxisLength [px]    :  76.20511
Orientation [grad]      :  2.27367
Solidity                :  0.49767
Extent                  :  0.38673
Eccentricity            :  0.62315
Convex Area [px]        :  5035.50000
Filled Area [px]        :  2506.00000
'''

    if show:
        print('--- extracting standard geometric features...')

    # Center of gravity
    I, J = np.where(image.astype(bool))
    i_m = I.mean()
    j_m = J.mean()

    # Area
    area = I.size

    # Perimeter
    L = perimeter(image)

    # Roundness
    roundness = 4 * area * np.math.pi / L ** 2 if L > 0 else np.nan

    # 'EulerNumber'
    filled_region = img.binary_fill_holes(image)
    _, n_objects = img.label(image)
    _, n_holes = img.label(filled_region - image)
    euler = n_objects - n_holes

    # 'FilledArea'
    filled_area = filled_region.sum()

    # 'ConvexArea'
    c_area = convex_area(image)

    # 'EquivDiameter'
    equiv_diameter = np.math.sqrt(4 * area / np.math.pi)

    # 'Solidity'
    solidity = area / c_area if c_area > 0 else np.nan

    # Bounding box
    row_slice, col_slice = bbox(image)
    height = row_slice.stop - row_slice.start
    width = col_slice.stop - col_slice.start
    bbox_area = height * width

    # 'Extent'
    extent = area / bbox_area if bbox_area > 0 else np.nan

    # MajorAxisLength, MinorAxisLength, Orientation and Eccentricity
    if area > 0:
        mu = moments(image, centered=True, order=2)

        a = mu[2, 0] / mu[0, 0]
        b = mu[1, 1] / mu[0, 0]
        c = mu[0, 2] / mu[0, 0]

        # eigen values of inertia tensor
        l1 = (a + c) / 2 + np.sqrt(4 * b ** 2 + (a - c) ** 2) / 2
        l2 = (a + c) / 2 - np.sqrt(4 * b ** 2 + (a - c) ** 2) / 2

        major_axis_len = 4 * np.sqrt(l1)
        minor_axis_len = 4 * np.sqrt(l2)
        eccentricity = 0 if l1 == 0 else np.sqrt(1 - l2 / l1)
        orientation = .5 if a - \
            c == 0 else np.math.atan2(2 * b, (a - c)) / np.pi
        if orientation < 0:
            orientation = (orientation + 1) * 90
        else:
            orientation = (orientation - 1) * 90

        # Danielsson shape factor (see Danielsson, 1977)
        td = img.morphology.distance_transform_cdt(image)
        dm = td[image.astype(bool)].mean()
        danielsson = area / 9 / np.pi / dm**2
    else:
        major_axis_len = np.nan
        minor_axis_len = np.nan
        orientation = np.nan
        eccentricity = np.nan
        danielsson = np.nan

    geo_features = np.array([i_m,
                             j_m,
                             height,
                             width,
                             area,
                             L,
                             roundness,
                             danielsson,
                             euler,
                             equiv_diameter,
                             major_axis_len,
                             minor_axis_len,
                             orientation,
                             solidity,
                             extent,
                             eccentricity,
                             c_area,
                             filled_area])

    if labels:
        return np.array(geo_labels), geo_features
    return geo_features


class BasicGeoExtractor(FeatureExtractor):

    def transform(self, X):
        return np.array([basic_geo_features(x) for x in self._get_iterator(X)])

    def get_labels(self):
        return np.array(geo_labels)



__all__ = ['gabor_features', 'GaborExtractor']

import numpy as np
import itertools as it
import tqdm

from pybalu.base import FeatureExtractor

_log2 = np.log(2)
_log2_sq = _log2 * _log2
sqrt_log2 = np.sqrt(_log2)
_2pi = 2 * np.pi


def gabor_kernel(p, q, L, sx, sy, u0, alpha, M):
    sx2 = sx * sx
    sy2 = sy * sy
    c = (M + 1) / 2
    ap = alpha ** (-p)
    tq = np.pi * q / L
    cos_tq = np.cos(tq)
    sin_tq = np.sin(tq)
    f_exp = 2 * np.pi * 1j * u0

    X = ap * np.repeat(np.arange(M) - c, M).reshape(M, M)
    Y = X.T

    _X = X * cos_tq + Y * sin_tq
    _Y = Y * cos_tq - X * sin_tq

    f = np.exp(-.5 * (_X * _X / sx2 + _Y * _Y / sy2)) * np.exp(f_exp * _X)
    return f * ap / (2 * np.pi * sx * sy)


def gabor_features(image, region=None, *, rotations=8, dilations=8, freq_h=2, freq_l=.1, mask=21, show=False, labels=False):
    '''\
    gabor_features(image, region=None, *, rotations=8, dilations=8, freq_h=2, freq_l=.1, mask=21, show=False, labels=False)

    (TODO)

    Parameters
    ----------

    Returns
    -------

    See Also
    --------

    Examples
    --------
    '''

    if mask % 2 == 0:
        raise ValueError(
            "`mask` value must be an odd positive integer, not '{mask}'")

    if region is None:
        region = np.ones_like(image)

    if show:
        print('--- extracting Gabor features...')

    alpha = (freq_h / freq_l) ** (1 / (dilations - 1))
    sx = sqrt_log2 * (alpha + 1) / (2 * np.pi * freq_h * (alpha-1))
    sy = sqrt_log2 - (2*_log2 / (_2pi * sx * freq_h))**2 /\
                     (_2pi * np.tan(np.pi / (2 * rotations)) *
                      (freq_h - 2 * np.log(1/4/np.pi**2/sx**2/freq_h)))
    u0 = freq_h

    k = np.where(region.astype(bool))
    N, M = image.shape

    g = np.zeros((dilations, rotations))
    size_out = image.shape + np.repeat(mask, 2) - 1
    Iw = np.fft.fft2(image, size_out)
    n1 = (mask + 1) // 2

    for p, q in it.product(range(dilations), range(rotations)):
        f = gabor_kernel(p, q, rotations, sx, sy, u0, alpha, mask)
        Ir = np.real(np.fft.ifft2(Iw * np.fft.fft2(np.real(f), size_out)))
        Ii = np.real(np.fft.ifft2(Iw * np.fft.fft2(np.imag(f), size_out)))
        Ir = Ir[n1:n1+N, n1:n1+M]
        Ii = Ii[n1:n1+N, n1:n1+M]
        Iout = np.sqrt(Ir*Ir + Ii*Ii)
        g[p, q] = Iout[k].mean()

    gmax = g.max()
    gmin = g.min()
    J = (gmax - gmin) / gmin
    features = np.hstack([g.ravel(), gmax, gmin, J])

    if labels:
        gabor_labels = np.hstack([
            [f'Gabor({p},{q})' for p, q in it.product(
                range(dilations), range(rotations))],
            ['Gabor-max', 'Gabor-min', 'Gabor-J']
        ])

        return gabor_labels, features

    return features


class GaborExtractor(FeatureExtractor):
    def __init__(self, *, rotations=8, dilations=8, freq_h=2, freq_l=.1, mask=21, show=False):
        self.show = show

        self.rotations = rotations
        self.dilations = dilations
        self.freq_h = freq_h
        self.freq_l = freq_l
        self.mask = mask

    def transform(self, X):
        params = self.get_params()
        params.update({'show': False})
        return np.array([gabor_features(x, **params) for x in self._get_iterator(X, desc='gabor')])

    def get_labels(self):
        return np.hstack([
            [f'Gabor({p},{q})' for p, q in it.product(
                range(self.dilations), range(self.rotations))],
            ['Gabor-max', 'Gabor-min', 'Gabor-J']
        ])



__all__ = ['morphoreg']
# pylint: disable=no-name-in-module
from pybalu.feature_extraction.geometric_utils import bw_perim
# pylint: enable=no-name-in-module
from skimage.morphology import binary_closing, remove_small_objects, remove_small_holes
import numpy as np
import warnings

_disk = np.ones((13, 13), dtype=np.uint8)
_disk[[0, 0, 0, 0, 1, 1, -2, -2, -1, -1, -1, -1],
      [0, 1, -2, -1, 0, -1, 0, -1, 0, 1, -2, -1]] = 0


def morphoreg(image, threshold=None):
    if threshold is None:
        R = image >= 0
    else:
        R = image > threshold

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore")
        clean = remove_small_objects(R, R.size // 100.0, connectivity=2)
        closed = binary_closing(clean, _disk)
        region = remove_small_holes(
            closed, R.size // 100.0, connectivity=2).astype(int)
        edge = bw_perim(region)

    return region, edge



__all__ = ['performance']


def performance(real, ideal):
    '''\
    performance(real, ideal)

    (TODO)
    '''
    real = real.squeeze()
    ideal = ideal.squeeze()

    if real.ndim != 1 or ideal.ndim != 1:
        raise ValueError(
            'Cannot compare classifications with more than one dimension')

    return (real == ideal).sum() / real.size



__all__ = ['jfisher']

import numpy as np

def jfisher(features, classification, p=None): 
    m = features.shape[1]
    
    norm = classification.ravel() - classification.min()
    max_class = norm.max() + 1
    
    if p is None:
        p = np.ones(shape=(max_class, 1)) / max_class
        
    # Centroid of all samples
    features_mean = features.mean(0)

    # covariance within class 
    cov_w = np.zeros(shape=(m, m))
    
    # covariance between classes
    cov_b = np.zeros(shape=(m, m))

    for k in range(max_class):
        ii = (norm == k)                                   # indices from class k
        class_features = features[ii,:]                    # samples of class k
        class_mean = class_features.mean(0)                # centroid of class k 
        class_cov = np.cov(class_features, rowvar=False)   # covariance of class k
        
        cov_w += p[k] * class_cov                          # within-class covariance
        
        dif = (class_mean - features_mean).reshape((m, 1))
        cov_b += p[k] * dif @ dif.T                        # between-class covariance
    try:
        return np.trace(np.linalg.inv(cov_w) @ cov_b)
    except np.linalg.LinAlgError:
        return - np.inf





import numpy as np
from pybalu.base import FeatureExtractor
__all__ = ['all_features', 'AllExtractor']


def all_features(image):
    return image.flatten()


class AllExtractor(FeatureExtractor):
    def transform(self, X):
        return np.array([all(x) for x in self._get_iterator(X)])



__all__ = ["structure"]

import numpy as np


def structure(classifier, *, train_data=None, test_data=None, train_classes=None, initialize=True,
              fit_method='fit', predict_method='predict', classifier_opts=None):
    '''\
    structure(classifier, train_data=None, test_data=None, train_classes=None, initialize=True,
              fit_method='fit', predict_method='predict, classifier_opts=None)

    Trains and/or classifies a set of data using the given classifier. Performs either or both actions
    depending on the given keyword arguments.

    Parameters 
    ----------
    classifier: Class or instance of a Class that defines train and predict methods
        The classifier to train or predict with. It can be a reference to the class or an instance of a class
        that defines two methods:
            - The `fit` method should accept a `train_data` and `train_classes` as input and should train the 
              classifier.
            - The `predict` method should accept `test_data` as input and return a numerical array-like.
        If an instance of a classifier is given, `initialize` parameter should be set to False.
        If a classifier Class is given, then it will be instanced with `classifier_opts` as initializer values.
    train_data: numerical 2 dimensional ndarray, optional
        Corresponds to the training data features. If none are given, it is assumed that the classifier is
        an instance of a classifier that is already trained. If this parameter is not None and no `train_classes`
        are given, a ValueError is thrown. 
    test_data: numerical 2 dimensional ndarray, optional
        Corresponds to the testing features. This will be given to the `predict` method of the given classifier.
    train_classes: integer ndarray, optional
        Corresponds to the classification of the given `train_data`. It is given to the `fit` method on the 
        classifier. Must be given if `test_data` is not None.
    initialize: bool, optional
        Wether the given `classifier` should be initialized or not. If `classifier` is an instance of a classifier, 
        this value should be set to False. In case `train_data` is None, this value is ignored, and `classifier` is
        supposed to be a trained instance. Default value is True, ie. 'classifier' is interpreted by default as a
        Class object and not an instance. 
    fit_method: string, optional
        The name of the training method on the classifier. Default is 'fit'.
    predict_method: string, optional
        The name of the classifying method on the classifier. Default is 'predict'.
    classifier_opts: dictionary, optional
        Represents the keyword arguments used to instantiate the classifier. Only useful if `classifier` is a Class
        object and `initialize` is set to True.

    Returns
    -------
    predicted_values: ndarray
        Represents the predicted classes of `test_data`. If no testing is performed, this will be an empty array.
    classifier_instance: instance of a classifier
        If `classifier` was an instance, this returns the same object. If `classifier` was a classifier Class object,
        an instance of this class is returned. If `train_data` was not None, then this object will be trained with 
        the given data.
    Examples
    --------
    ( TODO )
    '''

    if not hasattr(classifier, fit_method) or not hasattr(classifier, predict_method):
        raise ValueError(
            f"`classifier` must define methods `{fit_method}` and `{predict_method}` in order to be considered a valid classifier"
        )

    classifier_opts = classifier_opts or dict()

    if initialize:
        classifier_instance = classifier(**classifier_opts)
    else:
        classifier_instance = classifier

    if train_data is not None:
        if train_classes is None:
            raise ValueError(
                "`train_features` given but no `classification` given to train classifier"
            )

        classifier_instance = classifier(**classifier_opts)
        getattr(classifier_instance, fit_method)(train_data, train_classes)

    if test_data is not None:
        prediction = getattr(classifier_instance, predict_method)(test_data)
        return np.array(prediction), classifier_instance

    return np.empty(0), classifier_instance



__all__ = ['stratify']

import numpy as np

def stratify(labels, portion):

    if portion < 0 or portion > 1:
        raise ValueError(f'portion must be a value between 0 and 1, not {portion}')
    
    labels = labels.squeeze()
    np.unique(labels, return_counts=True)

    idxs_chosen = []
    idxs_not_chosen = []
    for cls, count in zip(*np.unique(labels, return_counts=True)):
        useful_idxs = np.where(labels == cls)[0]
        np.random.shuffle(useful_idxs)
        cls_portion = int(np.round(count * portion))
        idxs_chosen.append(useful_idxs[:cls_portion])
        idxs_not_chosen.append(useful_idxs[cls_portion:])


    return np.hstack(idxs_chosen), np.hstack(idxs_not_chosen)




__all__ = ['noorm_cooc_mtrx', 'cooc_features']

cimport numpy as np
cimport cython
import numpy as np

@cython.wraparound(False)
@cython.boundscheck(False)
def norm_cooc_mtrx(np.ndarray I, np.ndarray[np.int_t, ndim=2] R, int d):
    
    cdef int i, j, v1, v2, h_tot, v_tot, d1_tot, d2_tot
    cdef np.ndarray[np.double_t, ndim=3] co_occ
    cdef np.ndarray[np.int_t, ndim=2] V
    cdef np.ndarray[np.int_t, ndim=1] ii, jj, tot
    
    # normalize for only 8 pixel values and get rid of values not in R
    V = np.zeros(shape=(I.shape[0] + 2*d, I.shape[1] + 2*d), dtype=int)
    V[d:I.shape[0]+d, d:I.shape[1]+d] = np.floor(I / 32).astype(int)
    ii, jj = np.where(R != 1)
    V[ii+d, jj+d] = -1

    co_occ = np.zeros(shape=(4, 8, 8), dtype=float)
    tot = np.zeros(4, dtype=int)
    
    for i in range(d, I.shape[0]):
        for j in range(d, I.shape[1]):
            v1 = V[i, j] 
            if v1 < 0: continue
            # vertical
            v2 = V[i+d, j]
            if v1 >= 0 and v2 >= 0:
                co_occ[0, v1, v2] += 1
                co_occ[0, v2, v1] += 1
                tot[0] += 2
            # diagonal 1
            v2 = V[i+d, j-d]
            if v1 >= 0 and v2 >= 0:
                co_occ[1, v1, v2] += 1
                co_occ[1, v2, v1] += 1
                tot[1] += 2
            # horizontal
            v2 = V[i, j+d]
            if v1 >= 0 and v2 >= 0:
                co_occ[2, v1, v2] += 1
                co_occ[2, v2, v1] += 1
                tot[2] += 2
            # diagonal 2
            v2 = V[i+d, j+d]
            if v1 >= 0 and v2 >= 0:
                co_occ[3, v1, v2] += 1
                co_occ[3, v2, v1] += 1
                tot[3] += 2
    return co_occ / tot.reshape(4, 1, 1)
    

@cython.wraparound(False)
@cython.boundscheck(False)
def cooc_features(np.ndarray[np.double_t, ndim=2] P):
    
    cdef N, i, j

    N = 8
    
    Pij = P.ravel()
    pxi = P.sum(1)
    pyj = P.sum(0)
    ux = pxi.mean()
    uy = pyj.mean()
    sx = pxi.std(ddof=1)
    sy = pyj.std(ddof=1)
    
    ii, jj = np.where(P >= 0)
    dif_ij = ii - jj
    dif_ij_sq = dif_ij * dif_ij
    
    pxy1 = np.zeros(N * 2 - 1)
    pxy2 = np.zeros(N)
     
    for i in range(P.shape[0]):
        for j in range(P.shape[1]):
            pxy1[i+j] += P[i, j]
            pxy2[abs(i - j)] += P[i, j]
    
    # angular second moment (??)
    energy = Pij @ Pij
    
    # simplified method of calculating -> same as the paper
    contrast = dif_ij_sq @ Pij
    
    # holy f*ck what is this??
    correlation = (((ii+1) * (jj+1)) @ Pij - ux * uy * N ** 2) / sx / sy
    
    # this is the same as contrast (??)
    sum_of_sq = contrast

    # same as paper!
    inverse_dif_moment =  Pij @ (1 / (dif_ij_sq + 1))
    
    sum_avg = (ii+jj+2) @ Pij
    
    sum_entropy = - pxy1 @ np.log(pxy1, where=pxy1>0)
    
    sum_var = (np.arange(2, N*2+1) - sum_entropy) @ pxy1
    
    # same as paper!
    entropy = - Pij @ np.log(Pij, where=(Pij>0))
    
    # Difference Variance
    difference_var = pxy2.var(ddof=1)
    
    # Difference Entropy
    difference_entropy = -pxy2 @ np.log(pxy2, where=pxy2>0)
    
    pxipyj = pxi[ii] * pyj[jj]
    pxipyj_log = np.log(pxipyj, where=pxipyj>0)

    HXY = entropy
    HXY1 = -Pij @ pxipyj_log
    HXY2 = -pxipyj @ pxipyj_log
    HX = -pxi @ np.log(pxi, where=pxi>0)
    HY = -pyj @ np.log(pyj, where=pyj>0)
    
    # Measures of correlation
    corr_1 =  (HXY - HXY1) / max(HX, HY) if HX > 0 or HY > 0 else 0
    corr_2 = 1 - np.exp(-2*(HXY2 - HXY))
    
    # Maximal Correlation Coefficient
    Q = np.divide(P, pyj, where=pyj>0) @ np.divide(P, pyj, where=pxi>0)
    eigQ = np.linalg.eigvalsh(Q)
    max_corr_coef = eigQ[7]
    
    return np.array([
        energy,
        contrast,
        correlation,
        sum_of_sq,
        inverse_dif_moment,
        sum_avg,
        sum_var,
        sum_entropy,
        entropy,
        difference_var,
        difference_entropy,
        corr_1,
        corr_2,
        max_corr_coef
    ])


__all__ = ['fourier_features', 'FourierExtractor']

import numpy as np
import itertools as it
from PIL import Image

from pybalu.base import FeatureExtractor


def fourier_features(image, region=None, *, vresize=64, hresize=64, vfreq=2, hfreq=2, show=False, labels=False):
    '''\
fourier_features(image, region=None, *, vresize=64, hresize=64, vfreq=2, hfreq=2, show=False, labels=False)

Return an array of with the fourier features extracted from an image given resize and frequency
values.

Parameters
----------
image : a numpy 2 dimensional float array
    It represents a grayscale image or just one dimension of color (eg: green channel)
region : a numpy 2 dimensional array, optional
    A ndarray of the same dimensions as `image` of boolean type. All pixels not set to
    true on this array, will be set to 0 on `image` before resizing and sampling.
    if not given, no pixels will be modified before resizing and sampling.
vresize: int, optional
    `image` will be resized to this height before sampling. Default is 64
hresize: int, optional
    `image` will be resized to this width before sampling. Default is 64
vfreq: int, optional
    Vertical sampling frequency. Default is 2
hfreq: int, optional
    Horizontal sampling frequency. Default is 2
show : bool, optional
    Wether to print or not messages during execution
labels : bool, optional
    Wether to return a second array that contains the label of each value. 

Returns
-------
labels: ndarray, optional
    A one dimensional string ndarray that contains the labels to each of the features.
    This array is only returned if `labels` is set to True.
features: ndarray
    A float ndarray that contains 2 * `vfreq` * `hfreq` features extracted from `image`
    after its resize.

See Also
--------
clp : <Not Implemented Yet>
dct : <Not Implemented Yet>
fourier_des : Extracts the fourier descriptors of an image.
gabor : ( TODO )
haralick : Extracts 28 haralick texture features from an image.
lbp : ( TODO )

Examples
--------
Load an image on its grayscale representation, then proceed to get its features:

>>> from pybalu.feature_extraction import fourier_features
>>> from pybalu.io import imread
>>> img = imread('testimg.png', flatten=True) # to grayscale
>>> features = fourier_features(img)

Print a binary image features:

>>> from pybalu.io import print_features
>>> labels, features = fourier_features(img, labels=True)
>>> print_features(labels, features)
Fourier Abs (1, 1)      :  6948.20449
Fourier Abs (1, 2)      :  966.09909
Fourier Abs (2, 1)      : -148.03935
Fourier Abs (2, 2)      :  92.26226
Fourier Ang (1, 1) [rad]: -0.13091
Fourier Ang (1, 2) [rad]:  0.07754
Fourier Ang (2, 1) [rad]:  0.12337
Fourier Ang (2, 2) [rad]:  0.01847
'''

    if region is None:
        region = np.ones_like(image)

    I = image.astype(float)
    I[region == 0] = 0

    v_half = round(vresize / 2) + 1
    h_half = round(hresize / 2) + 1

    if show:
        print('--- extracting Fourier features...')

    img_resize = _imresize(I, (vresize, hresize))
    img_fourier = np.fft.fft2(img_resize)
    img_abs = np.abs(img_fourier)
    img_angle = np.angle(img_fourier)

    F = _imresize(img_abs[:v_half, :h_half], (vfreq, hfreq))
    A = _imresize(img_angle[:v_half, :h_half], (vfreq, hfreq))

    features = np.hstack([F.ravel(), A.ravel()]).astype(float)
    if labels:
        fourier_labels = np.zeros(vfreq * hfreq * 2, dtype='<U28')
        fourier_labels[:vfreq * hfreq] = [f'Fourier Abs ({i}, {j})' for i, j in
                                          it.product(range(1, vfreq+1), range(1, hfreq+1))]
        fourier_labels[vfreq * hfreq:] = [f'Fourier Ang ({i}, {j}) [rad]' for i, j in
                                          it.product(range(1, vfreq+1), range(1, hfreq+1))]
        return fourier_labels, features
    return features


def _imresize(arr, size):
    """Used to keep compatibility with deprecated scipy function `imresize`"""
    return np.array(Image.fromarray(arr).resize(size))


class FourierExtractor(FeatureExtractor):
    def __init__(self, *, vresize=64, hresize=64, vfreq=2, hfreq=2):
        self.vresize = vresize
        self.hresize = hresize
        self.vfreq = vfreq
        self.hfreq = hfreq

    def transform(self, X):
        params = self.get_params()
        return np.array([fourier_features(x, **params) for x in self._get_iterator(X)])

    def get_labels(self):
        labels = np.zeros(self.vfreq * self.hfreq * 2, dtype='<U28')
        labels[:self.vfreq * self.hfreq] = [f'Fourier Abs ({i}, {j})' for i, j in
                                            it.product(range(1, self.vfreq+1), range(1, self.hfreq+1))]
        labels[self.vfreq * self.hfreq:] = [f'Fourier Ang ({i}, {j}) [rad]' for i, j in
                                            it.product(range(1, self.vfreq+1), range(1, self.hfreq+1))]
        return labels



__all__ = ['basic_int_features', 'BasicIntExtractor']

import numpy as np
import scipy.stats
# pylint: disable=no-name-in-module
from pybalu.img_processing import fst_deriv, snd_deriv
from .geometric_utils import bw_perim
# pylint: enable=no-name-in-module
from pybalu.base import FeatureExtractor

int_labels = ['Intensity Mean',
              'Intensity StdDev',
              'Intensity Kurtosis',
              'Intensity Skewness',
              'Mean Laplacian',
              'Mean Boundary Gradient']


def basic_int_features(image, region=None, *, mask=15, show=False, labels=False):
    '''\
basic_int_features(image, region=None, *, mask=15, show=False, labels=False)

Return an array of with the basic intensity features of a grayscale image.

Parameters
----------
image : a numpy 2 dimensional float array
    It represents a grayscale image or just one dimension of color (eg: green channel)
region : a numpy 2 dimensional array, optional
    A ndarray of the same dimensions as `image` of boolean type. Only the pixels that are
    True in this array will be analyzed. If it is not set, as a default the whole image
    will be used.
mask: int, optional
    Gauss mask value for gradient computation. Default is 15
show : bool, optional
    Wether to print or not messages during execution
labels : bool, optional
    Wether to return a second array that contains the label of each value. 

Returns
-------
labels: ndarray, optional
    A one dimensional string ndarray that contains the labels to each of the features.
    This array is only returned if `labels` is set to True.
features: ndarray
    A float ndarray that contains all 6 intensity features extracted from pixels in 
    `image` contained in the `region` given.

See Also
--------
clp : <Not Implemented Yet>
dct : <Not Implemented Yet>
fourier : Extracts the fourier features of an image.
gabor : <Not Implemented Yet> 
haralick : Extracts 28 haralick texture features from an image.
lbp : <Not Implemented Yet>

Examples
--------
Load an image on its grayscale representation, then proceed to get its features:

>>> from pybalu.feature_extraction import basic_int_features
>>> from pybalu.img_processing import segbalu
>>> from pybalu.io import imread
>>> img = imread('testimg.png', flatten=True) # to grayscale
>>> region, _, _ = segbalu(img)
>>> features = basic_int_features(img, region)

Print image features:

>>> from pybalu.io import print_features
>>> labels, features = basic_int_features(img, region, labels=True)
>>> print_features(labels, features)
Intensity Mean        :  224.86193
Intensity StdDev      :  66.68996
Intensity Kurtosis    :  6.26327
Intensity Skewness    : -2.15027
Mean Laplacian        : -23.24741
Mean Boundary Gradient:  48.57239
'''
    if region is None:
        region = np.ones(shape=image.shape, dtype=int)
    if show:
        print('--- extracting basic intensity features...')

    r_perim = bw_perim(region, 4).astype(bool)
    region = region.astype(bool)

    image = image.astype(float)

    im1, _, _ = fst_deriv(image, mask=mask)
    im2 = snd_deriv(image)

    if not region.all():
        boundary_gradient = np.abs(im1[r_perim]).mean()
    else:
        boundary_gradient = -1

    useful_img = image[region]

    intensity_mean = useful_img.mean()
    intensity_std = useful_img.std(ddof=1)
    intensity_kurtosis = scipy.stats.kurtosis(useful_img, fisher=False)
    intensity_skewness = scipy.stats.skew(useful_img)
    mean_laplacian = im2[region].mean()

    int_features = np.array([intensity_mean,
                             intensity_std,
                             intensity_kurtosis,
                             intensity_skewness,
                             mean_laplacian,
                             boundary_gradient])

    if labels:
        return np.array(int_labels), int_features
    return int_features


class BasicIntExtractor(FeatureExtractor):
    def __init__(self, *, mask=15):
        self.mask = mask

    def transform(self, X):
        params = self.get_params()
        return np.array([basic_int_features(x, **params) for x in self._get_iterator(X)])

    def get_labels(self):
        return np.array(int_labels)



__all__ = ['segbalu', 'SegBaluSegmentator']

import numpy as np
from pybalu.base import ImageProcessor
from skimage.filters import threshold_otsu
from .rgb2hcm import rgb2hcm
from .morphoreg import morphoreg


def segbalu(image, p=-.05):
    '''\
    segbalu(image, p=-.05)

    Segmentation of an object with homogeneous background.

    Parameters
    ----------
    image: 3 dimensional ndarray
        The RGB input image.
    p: float between -1 and 1, optional
        A positive value is used to dialte the segmentation. A negative erodes.
        default value is -0.05

    Returns
    -------
    region: 2 dimensional ndarray
        Binary image of the object
    edges: 2 dimensional ndarray
        Binary image of the edges of the object
    hcm: 2 dimensional ndarray
        high contrast grayscale representation of input image

    See Also
    --------
    (TODO)

    Examples
    --------
    (TODO)
    '''
    hcm = rgb2hcm(image.astype('double'))
    threshold = threshold_otsu(hcm)
    region, edge = morphoreg(hcm, threshold + p)
    return region, edge, hcm


class SegBaluSegmentator(ImageProcessor):
    def __init__(self, *, p=-.05, returns='all', show=False):
        self.show = show

        self.p = p
        self.returns = returns
        if returns == 'all':
            self.idx = slice(0, 3)
        elif returns in ['region', 'edges', 'hcm']:
            self.idx = ['region', 'edges', 'hcm'].index(returns)
        else:
            raise ValueError(
                f"Invalid value for <returns>: '{returns}'"
            )

    def transform(self, X):
        return np.array([segbalu(x, self.p)[self.idx] for x in self._get_iterator(X, desc=f'segbalu_{self.returns}')])



__all__ = ['haralick_features', 'HaralickExtractor']

import numpy as np
from collections import Sequence

from pybalu.base import FeatureExtractor
# pylint: disable=no-name-in-module
from .haralick_utils import norm_cooc_mtrx, cooc_features
# pylint: enable=no-name-in-module


def haralick_features(image, region=None, *, distance=3, show=False, labels=False):
    '''\
haralick_features(image, region=None, *, distance=3, show=False, labels=False)

Return an array of with the mean and range of 14 different measures over the
coocurrence matrix of a grayscale image.

Parameters
----------
image : a numpy 2 dimensional float array
    It represents a grayscale image or just one dimension of color (eg: green channel).
    Its values should range from 0 to 255.
region : a numpy 2 dimensional array, optional
    A ndarray of the same dimensions as `image` of boolean type. Only the pixels that are
    True in this array will be analyzed. If it is not set, as a default the whole image
    will be used.
distance : (int, array_like), optional
    If `distance` is an int, it defines the distance for which to analyze pixel coocurrence.
    If `distance` is an array_like, it is expected to be a sequence of ints, all of which
    define the distance for which to analyze pixel coocurrence. All features are then extracted
    for each of these distances.
    Default value is 3
show : bool, optional
    Wether to print or not messages during execution
labels : bool, optional
    Wether to return a second array that contains the label of each value. 

Returns
-------
labels: ndarray, optional
    A one dimensional string ndarray that contains the labels to each of the features.
    This array is only returned if `labels` is set to True.
features: ndarray
    A float ndarray that contains the 28 * len(`distance`) values of various measures
    extracted from the coocurrence matrix of `image`.

See Also
--------
clp : <Not Implemented Yet>
dct : <Not Implemented Yet>
fourier : Extracts the fourier features of an image.
gabor : <Not Implemented Yet> 
lbp : <Not Implemented Yet>
Examples
--------
Load an image and get its grayscale representation, then proceed to get its features
for 3 distances:

>>> from pybalu.feature_extraction import haralick_features
>>> from pybalu.io import imread
>>> img = imread('testimg.png')
>>> features = haralick_features(img, distance=[3, 4, 5])

Print the haralick_features features for distance 12:

>>> from pybalu.io import print_features
>>> labels, features = haralick_features(img, distance=12, labels=True)
>>> print_features(labels, features)
Tx 1 , d 12 (mean) :  0.43780
Tx 2 , d 12 (mean) :  13.47477
Tx 3 , d 12 (mean) :  72.48584
Tx 4 , d 12 (mean) :  13.47477
Tx 5 , d 12 (mean) :  0.69266
Tx 6 , d 12 (mean) :  4.77283
Tx 7 , d 12 (mean) :  3.66382
Tx 8 , d 12 (mean) :  1.10901
Tx 9 , d 12 (mean) :  1.33820
Tx 10, d 12 (mean) :  0.05783
Tx 11, d 12 (mean) :  0.91347
Tx 12, d 12 (mean) : -0.03575
Tx 13, d 12 (mean) :  0.04514
Tx 14, d 12 (mean) :  0.05285
Tx 1 , d 12 (range):  0.49309
Tx 2 , d 12 (range):  15.46938
Tx 3 , d 12 (range):  107.59946
Tx 4 , d 12 (range):  15.46938
Tx 5 , d 12 (range):  0.79220
Tx 6 , d 12 (range):  4.79386
Tx 7 , d 12 (range):  3.68360
Tx 8 , d 12 (range):  1.12357
Tx 9 , d 12 (range):  1.36838
Tx 10, d 12 (range):  0.07199
Tx 11, d 12 (range):  0.96111
Tx 12, d 12 (range):  0.13119
Tx 13, d 12 (range):  0.16454
Tx 14, d 12 (range):  0.20168
'''
    image = image.astype('double')

    if region is None:
        region = np.ones_like(image).astype(int)

    if show:
        print('--- extracting Haralick texture features...')

    if not isinstance(distance, Sequence):
        distance = [distance]
    dseq = np.array(distance, dtype=int).ravel()

    features = []
    label_list = []
    for d in dseq:
        cooc = norm_cooc_mtrx(image, region, d)
        feats = np.vstack([cooc_features(P) for P in cooc])
        features.append(np.hstack([feats.mean(0), np.abs(feats).max(0)]))
        if labels:
            label_list.extend(
                [f"Tx {i:<2d}, d {d:<2d} (mean)" for i in range(1, 15)])
            label_list.extend(
                [f"Tx {i:<2d}, d {d:<2d} (range)" for i in range(1, 15)])

    haralick_features = np.hstack(features)

    if labels:
        return np.array(label_list), haralick_features
    return haralick_features


class HaralickExtractor(FeatureExtractor):
    def __init__(self, *, distance=3, show=False):
        self.show = show
        self.distance = distance

    def transform(self, X):
        params = self.get_params()
        params.update({'show': False})
        return np.array([haralick_features(x, **params) for x in self._get_iterator(X, desc='haralick')])

    def get_labels(self):
        if isinstance(self.distance, Sequence):
            distance = self.distance
        else:
            distance = [self.distance]
        labels = []
        for d in distance:
            labels.extend(
                [f"Tx {i:<2d}, d {d:<2d} (mean)" for i in range(1, 15)])
            labels.extend(
                [f"Tx {i:<2d}, d {d:<2d} (range)" for i in range(1, 15)])
        return np.array(labels)



__all__ = ['pca']
import numpy as np
from scipy.linalg import eigh

def pca(features, *, n_components=0, energy=0):
    '''\
    pca(features, n_components=0, energy=0)
    
    Principal component analysis

    Parameters
    ----------
    features: numerical 2 dimensional ndarray
        Corresponds to the sample features over which to perform PCA.
    n_components: positive integer, optional
        Number of selected components. If both `n_components` and `energy` are not given,
        `n_components` is set to the original number of features.
    energy: float in range [0, 1], optional
        If `n_components` is not set and `energy` is greater than 0, `n_components` is set
        to the lowest possible number that fulfills the condition
            sum(stdev[:n_components])) / sum(stdev) >= energy.
        where `stdev` corresponds to the standard deviation of each principal component.

    Returns
    -------
    p_components: float ndarray 
        The `n_components` principal components of `features`, sorted according to decreasing
        `stdev`.
    stdev: float ndarray
        The standard deviation of each principal component. One dimensional array 
        of size `n_components`.
    transform: float ndarray
        The transformation matrix for obtaining the principal components.
    feat_mean: float ndarray
        The mean of each feature in `features`.
    new_features: float ndarray
        Of the same shape as `features`. Corresponds to the new features calculated
        from the selected principal components. It corresponds to performing the 
        following operation: `p_components` @ `selected_transform`.T + `feat_mean`

    Examples
    --------
    Reduce a grayscale image to 30 principal components and display together with original
    
    >>> from pybalu.io import imread
    >>> from pybalu.feature_transformation import pca
    >>> import matplotlib.pyplot as plt
    >>> 
    >>> img = imread('cameraman.png', flatten=True)
    >>> _, _, _, _, new_img = pca(img, n_components=30)
    >>> plt.subplot(121); plt.imshow(img, cmap='gray'); plt.subplot(122); plt.imshow(new_image, cmap='gray')
    >>> plt.show()


    Reduce the number of features

    (TODO)

    '''
    N = features.shape[1]

    if not isinstance(n_components, int):
        raise ValueError(f'`n_components` must be a positive int, not {n_components}')
    if n_components <= 0:
        if energy == 0:
            raise ValueError('Either `n_components` or `energy` must be set')
        if not 0 < energy < 1:
            raise ValueError(f'`energy` must be a value between 0 and 1, not {energy}')
        # energy is a valid value
        n_components = N
    else:
        # n_components is a valid value
        energy = 0
    
    feat_mean = features.mean(0)
    X0 = features - feat_mean
    Cx = np.cov(X0, rowvar=False)

    if energy > 0:
        # calculate ALL eigenvectors
        eig_vals, eig_vecs = np.linalg.eigh(Cx) 
        # reverse them into descending value 
        _lambda = eig_vals[::-1]
        transform = eig_vecs[:,::-1]

        # calculate the number of selected components
        energy_eig = np.tril(np.ones((N, N))) @ _lambda / _lambda.sum()
        n_components = np.where(energy_eig > energy)[0][0] + 1
        transform = transform[:, :n_components]
        _lambda = _lambda[:n_components]

    else:
        # calculate only the wanted number of eigenvectors
        eig_vals, eig_vecs = eigh(Cx, eigvals=(N-n_components, N-1))
        # reverse them into descending value 
        _lambda = eig_vals[::-1]
        transform = eig_vecs[:,::-1]

    p_components = X0 @ transform
    new_features = p_components @ transform.T + feat_mean

    return p_components, _lambda, transform, feat_mean, new_features


__all__ = ['sep_into_bins']

cimport cython
cimport numpy as np
import numpy as np

@cython.boundscheck(False)
@cython.wraparound(False)
def sep_into_bins(double[:] vals, long[:] bins, int n_bins):
    cdef double[:] out
    cdef int i, b, N
    cdef float val
    
    N = bins.size
    out = np.zeros(n_bins)
    for i in range(N):
        out[bins[i]] += vals[i]
        
    return np.asarray(out)


__all__ = ['crossval']
import numpy as np
from pybalu.classification import structure
from .performance import performance
from scipy.stats import t, norm


def crossval(features, classification, classifier, n_folds, *, confidence=.95, classifier_opts=None, strat=False):
    '''\
    crossval(features, classification, classifier, n_folds, *, confidence=.95, options=None, strat=False)

    n-fold Cross Validation in n groups of given samples and classification 
    according to the given classifier.

    Parameters 
    ----------
    features: numerical 2 dimensional ndarray
        Corresponds to the sample features over which to perform cross-validation.
    classification: integer ndarray
        Corresponds to the classification of the given samples.
    classifier: Class that defines train and predict methods
        The classifier to train or predict with. It can be a reference to a class that defines two methods:
            - The `fit` method should accept a subset of `features` and `classification` as input and 
            should train the classifier.
            - The `predict` method should accept a subset of `features` as input and return a numerical array-like.
        The classifier will be instanced with `classifier_opts` as initializer values.
    n_folds: integer
        Number of groups (folders) of the cross-validation
    condifence: float in range [0, 1]
        The probability of the confidence interval.
    classifier_opts: dictionary, optional
        Represents the keyword arguments used to instantiate the classifier.
    strat: bool, optional
        If set to True, the classes are stratified within the folds. Default value is False

    Returns
    -------
    performance: float in range [0, 1]
        The mean performance of classifier.
    confidence_interval: (float, float)
        Confidence intervals for `confidence` probability.

    Examples
    --------
    ( TODO )
    '''
    N = features.shape[0]
    if N != classification.size:
        raise ValueError(
            f'Dim 0 of `features` must be equal to `classification` size. ({N} != {classification.size})')

    # shuffle features and then sort features by classification
    perm = np.random.permutation(np.arange(N))
    features = features[perm]
    classification = classification[perm].squeeze()
    idx_sort = np.argsort(classification)
    features = features[idx_sort]
    classification = classification[idx_sort]

    classes, counts = np.unique(classification, return_counts=True)
    n_cls = len(classes)

    if strat and any(counts < n_folds):
        raise ValueError(
            'Some classes have less occurences than the number of folds!')

    if n_folds == 1:
        f_train = features
        f_test = features
        d_train = classification
        d_test = classification

        d_fitted = structure(classifier,
                             train_data=f_train,
                             test_data=f_test,
                             train_classes=d_train,
                             classifier_opts=classifier_opts)

        return performance(d_test, d_fitted), (0., 0.)
    if strat:
        # generate blocks
        cls_size = counts // n_folds
        starts = np.tril(np.ones((n_cls, n_cls), dtype=int), -1) @ counts
        ends = np.hstack([starts[1:], N])
        idx_vec = np.hstack([np.arange(s, s+v)
                             for s, v in zip(starts, cls_size)])
        diff = np.repeat(cls_size, cls_size)
        starts += cls_size * (n_folds - 1)
        first_blocks = [idx_vec + diff*v for v in range(n_folds-1)]
        last_block = [np.hstack([np.arange(s, e)
                                 for s, e in zip(starts, ends)])]
        blocks = first_blocks + last_block
    else:
        block_size = N // n_folds
        first_blocks = [np.arange(block_size*i, block_size*(i+1))
                        for i in range(n_folds-1)]
        last_block = [np.arange(block_size*(n_folds-1), N)]
        blocks = first_blocks + last_block

    pp = np.zeros(n_folds)
    for n, block in enumerate(blocks):
        other = np.delete(np.arange(N), block)
        f_train = features[block]
        f_test = features[other]
        d_train = classification[block]
        d_test = classification[other]
        d_fitted, _ = structure(classifier,
                                train_data=f_train,
                                test_data=f_test,
                                train_classes=d_train,
                                classifier_opts=classifier_opts)

        pp[n] = performance(d_test, d_fitted)

    pm = pp.mean()
    std = np.sqrt((pm * (1 - pm) / N))
    if n_folds > 20:
        return pm, norm.interval(confidence, loc=pm, scale=std)

    return pm, t.interval(confidence, n_folds - 1, loc=pm, scale=std)



__all__ = ['sfs', 'SFSWarning']

import numpy as np
import warnings
import tqdm
from pybalu.feature_analysis.score import score


class SFSWarning(Warning):
    pass


def sfs(features, classes, n_features, *, force=False, method='fisher', options=None, show=False):

    N, M = features.shape
    remaining_feats = set(np.arange(M))
    selected = list()
    curr_feats = np.zeros((N, 0))
    if options is None:
        options = dict()

    def _calc_score(i):
        feats = np.hstack([curr_feats, features[:, i].reshape(-1, 1)])
        return score(feats, classes, method=method, **options)

    if show:
        _range = tqdm.trange(
            n_features, desc='Selecting Features', unit_scale=True, unit=' features')
    else:
        _range = range(n_features)

    for _ in _range:
        new_selected = max(remaining_feats, key=_calc_score)
        selected.append(new_selected)
        remaining_feats.remove(new_selected)
        curr_feats = np.hstack(
            [curr_feats, features[:, new_selected].reshape(-1, 1)])

    return np.array(selected)



__all__ = ['perimeter', 'moments', 'convex_area', 'bbox', 'bw_perim']

cimport cython
cimport numpy as np
import numpy as np
import scipy.spatial

@cython.wraparound(False)
@cython.boundscheck(False)
cpdef float perimeter(np.ndarray[np.int_t, ndim=2] region):

    cdef int h = region.shape[0]
    cdef int w = region.shape[1]
    # first account for border pixels
    cdef float p = region[0].sum() + region[h-1].sum() + region[1:h-1,0].sum() + region[1:h-1,w-1].sum()
    
    cdef int i, j
    
    for i in range(1, h-1):
        for j in range(1, w-1):
            if region[i,j] == 0: continue
            #connectivity 4
            if region[i-1, j] == 0 or region[i+1, j] == 0:
                p += 1
                continue
            if region[i, j-1] == 0 or region[i, j+1] == 0:
                p += 1
                continue

            #connectivity 8
            if region[i-1, j-1] == 0 or region[i-1, j+1] == 0:
                p += .25
                continue
            if region[i+1, j-1] == 0 or region[i+1, j+1] == 0:
                p += .25
                continue
    return p

@cython.wraparound(False)
@cython.boundscheck(False)
cpdef np.ndarray[np.int_t, ndim=2] bw_perim(np.ndarray[np.int_t, ndim=2] region, int conn=4):
    if not (conn == 4 or conn == 8):
        raise Exception('Neighbourhood connectivity must be either 4 or 8, not ' + str(conn))
    
    cdef int h = region.shape[0]
    cdef int w = region.shape[1]

    cdef np.ndarray[np.int_t, ndim=2] out = np.zeros(shape=(h, w), dtype=int)
    
    cdef int i, j
    
    for i in range(1, h-1):
        for j in range(1, w-1):
            if not region[i,j]: continue
            #connectivity 4
            if not region[i-1, j] or not region[i+1, j]:
                out[i, j] = 1
                continue
            if not region[i, j-1] or not region[i, j+1]:
                out[i, j] = 1
                continue
            if conn == 4: continue
            #connectivity 8
            if not region[i-1, j-1] or not region[i-1, j+1]:
                out[i, j] = 1
                continue
            if not region[i+1, j-1] or not region[i+1, j+1]:
                out[i, j] = 1
                continue
    return out


@cython.wraparound(False)
@cython.boundscheck(False)
def moments(np.ndarray[np.int_t, ndim=2] region, centered=False, order=3):
    cdef int i, j
    cdef np.ndarray[np.double_t, ndim=2] m, I_, J_
    
    I, J = np.where(region == 1)
    I_ = np.ones((order+1, I.size))
    J_ = np.ones((order+1, J.size))
    I_[1,:] = I
    J_[1,:] = J

    if centered:
        I_[1,:] -= I.mean()
        J_[1,:] -= J.mean()
        
    for i in range(2, order + 1):
        I_[i,:] = I_[i-1,:] * I_[1,:]
        J_[i,:] = J_[i-1,:] * J_[1,:]

    m = np.zeros(shape=(order+1,order+1), dtype=float)
    for i in range(order + 1):    
        for j in range(order + 1):
            m[i,j] = I_[i,:] @ J_[j,:]
    return m


@cython.wraparound(False)
@cython.boundscheck(False)
cpdef bbox(np.ndarray[np.int_t, ndim=2] region):
    
    cdef int n, m, i, j, min_i, max_i, min_j, max_j
    
    max_i = max_j = 0
    n = min_i = region.shape[0]
    m = min_j = region.shape[1]
    
    for i in range(n):
        for j in range(m):
            if region[i, j] == 1:
                if i < min_i: min_i = i
                if i > max_i: max_i = i
                if j < min_j: min_j = j
                if j > max_j: max_j = j
    
    return slice(min_i, max_i + 1), slice(min_j, max_j + 1)


@cython.wraparound(False)
@cython.boundscheck(False)
def convex_area(np.ndarray[np.int_t, ndim=2] region):
    cdef np.ndarray[np.int_t, ndim=1] ii, jj
    cdef np.ndarray[np.double_t, ndim=1] X, Y
    cdef np.ndarray[np.double_t, ndim=2] dx, dy
    cdef list hull
    cdef int i, p, q, N
    cdef float val
    
    
    if region.size < 1: return 0

    ii, jj = np.where(bw_perim(region).astype(bool))
    dx = np.array([-0.5, -0.5, 0.5, 0.5]).reshape(4, 1)
    dy = np.array([-0.5, 0.5, -0.5, 0.5]).reshape(4, 1)
    X = (ii + dx).flatten()
    Y = (jj + dy).flatten()
    
    if X.size == 0:
        return 0

    hull = []
    
    area = 0
    i = 0
    p = 0
    q = 0
    
    while True:
        hull.append(p)
        q = (p + 1) % X.size;
        for i in range(X.size):
            val = (Y[i] - Y[p]) * (X[q] - X[i]) - (X[i] - X[p]) * (Y[q] - Y[i])
            if val < 0:
                q = i
        p = q
        if p == 0: 
            break
    hull.append(0)
    X, Y = X[hull], Y[hull]
    N = X.size
    return (X[:N-1] @ Y[1:] - X[1:] @ Y[:N-1]) / 2

@cython.wraparound(False)
@cython.boundscheck(False)
cdef inline tuple nnext_cw(tuple curr, tuple _from):
    cdef int i, j, ci, cj
    i, j = curr
    ci, cj = _from
    
    if i == ci:
        if j < cj:
            return i-1, j
        return i+1, j
    if j == cj:
        if i < ci:
            return i, j+1
        return i, j-1
    if i < ci:
        if j < cj:
            return i, cj
        return ci, j
    if j < cj:
        return ci, j
    return i, cj

@cython.wraparound(False)
@cython.boundscheck(False)
cpdef bw_boundaries(np.ndarray[np.int_t, ndim=2] R):
    cdef tuple s, b, p, c
    cdef Py_ssize_t i, j, N, M
    cdef list B
    cdef np.ndarray[np.int_t, ndim=2] _R

    N = R.shape[0]
    M = R.shape[1]
    _R = np.zeros((N+2, M+2), dtype=int)
    _R[1:N+1,1:M+1] = R
    
    s = None
    i = N
    j = 1
    
    while i > 0 and s is None:
        j = 1
        while j < M:
            if _R[i,j] == 1:
                s = (i,j)
                break
            j += 1
        i -= 1
    if j == 1:
        b = i + 2, j
    else:
        b = i + 1, j - 1
    B = [s]
    p = s
    c = nnext_cw(b, p)
    while c != s:
        if _R[c]:
            B.append(c)
            b = p
            p = c
        else:
            b = c
        c = nnext_cw(b, p)
    return np.array(B) - 1


__all__ = ['exsearch']

import numpy as np
from itertools import combinations
from pybalu.feature_analysis.score import score
import warnings

import tqdm


def choose(n, k):
    return int(np.math.factorial(n) / (np.math.factorial(n - k) * np.math.factorial(k)))


def exsearch(features, classes, n_features, *, method='fisher', options=None, show=False):
    if options is None:
        options = dict()

    tot_feats = features.shape[1]
    N = choose(tot_feats, n_features)

    if N > 10000:
        warnings.warn(
            f'Doing more than 10.000 iterations ({N}). This may take a while...')

    def _calc_score(ii):
        feats = features[:, ii]
        return score(feats, classes, method=method, **options)

    _combinations = combinations(range(tot_feats), n_features)

    if show:
        _combinations = zip(tqdm.trange(N,
                                        desc='Combinations checked',
                                        unit_scale=True,
                                        unit=' combinations'),
                            _combinations)

        _combinations = (ii for _, ii in _combinations)

    chosen_feats = max(_combinations, key=_calc_score)

    return np.array(chosen_feats)



# modules

__all__ = ["classification",
           "data_selection",
           "feature_analysis",
           "feature_extraction",
           "feature_selection",
           "feature_transformation",
           "img_processing",
           "io",
           "misc",
           "performance_eval"]

__version__ = '0.2.9'



import numpy as np
from pybalu.classification import structure
from pybalu.performance_eval import performance
from .sp100 import sp100
from .jfisher import jfisher

__all__ = ['score']


def score(features, classification, *, method='fisher', param=None):
    if param is None:
        dn = classification.max() - classification.min() + 1  # number of classes
        p = np.ones((dn, 1)) / dn
    else:
        p = param

    if method == 'mi':  # mutual information
        raise NotImplementedError()

    # maximal relevance
    elif method == 'mr':
        raise NotImplementedError()

    # minimal redundancy and maximal relevance
    elif method == 'mrmr':
        raise NotImplementedError()

    # fisher
    elif method == 'fisher':
        return jfisher(features, classification, p)

    elif method == 'sp100':
        return sp100(features, classification)

    else:
        return 0



__all__ = ['gupta_features', 'GuptaExtractor']

import numpy as np

from pybalu.base import FeatureExtractor
# pylint: disable=no-name-in-module
from .geometric_utils import bw_perim
# pylint: enable=no-name-in-module

_gupta_labels = ['Gupta-moment 1',
                 'Gupta-moment 2',
                 'Gupta-moment 3']


def gupta_features(image, *, show=False, labels=False):
    '''\
gupta_features(image, *, show=False, labels=False)

Return an array of with the 3 Gupta moments of a binary image.

Parameters
----------
image : a numpy 2 dimensional int array
    It represents a binary image. Non binary arrays will return nonsensical results.
show : bool, optional
    Wether to print or not messages during execution
labels : bool, optional
    Wether to return a second array that contains the label of each value. 

Returns
-------
labels: ndarray, optional
    A one dimensional string ndarray that contains the labels to each of the features.
    This array is only returned if `labels` is set to True.
features: ndarray
    A float ndarray that contains the 3 Gupta moments extracted from `image`.

See Also
--------
basic_geo : Extract the standard geometric features from a binary image.
fit_ellipse : <Not Implemented Yet>
flusser : Extract the four Flusser moments from binary image.
hugeo : Extract the seven Hu moments from a binary image.

Examples
--------
Load an image and get its binary representation, then proceed to get its features:

>>> from pybalu.feature_extraction import gupta_features
>>> from pybalu.img_processing import segbalu
>>> from pybalu.io import imread
>>> img = imread('testimg.png')
>>> binary_img, _, _ = segbalu(img)
>>> features = gupta_features(binary_img)

Print a binary image features:

>>> from pybalu.io import print_features
>>> labels, features = gupta_features(binary_img, labels=True)
>>> print_features(labels, features)
Gupta-moment 1:  0.41745
Gupta-moment 2: -0.13898
Gupta-moment 3:  2.25704
'''
    if show:
        print('--- extracting Gupta moments...')

    i_perim, j_perim = np.where(bw_perim(image, 4).astype(bool))
    im_perim = i_perim + j_perim * 1j
    ix = i_perim.mean()
    jx = j_perim.mean()
    centre = ix + jx * 1j
    z = np.abs(im_perim - centre)
    m1 = z.mean()

    mur1 = z - m1
    mur2 = mur1 * mur1
    mur3 = mur1 * mur2
    mur4 = mur2 * mur2

    mu2 = mur2.mean()
    mu3 = mur3.mean()
    mu4 = mur4.mean()

    F1 = (mu2 ** .5) / m1
    F2 = mu3 / (mu2 * (mu2 ** .5))
    F3 = mu4 / mu2 ** 2

    gupta_moments = np.array([F1, F2, F3])

    if labels:
        return np.array(_gupta_labels), gupta_moments
    return gupta_moments


class GuptaExtractor(FeatureExtractor):

    def transform(self, X):
        return np.array([gupta_features(x) for x in self._get_iterator(X)])

    def get_labels(self):
        return np.array(_gupta_labels)



# -*- coding: utf-8 -*-
#
import ast
from setuptools.extension import Extension
from setuptools import setup
import os
import sys
"""setuptools-based setup.py template for Cython projects.

Main setup for the library.

Usage as usual with setuptools:
    python setup.py build_ext
    python setup.py build
    python setup.py install
    python setup.py sdist

For details, see
    http://setuptools.readthedocs.io/en/latest/setuptools.html#command-reference
or
    python setup.py --help
    python setup.py --help-commands
    python setup.py --help bdist_wheel  # or any command
"""

#########################################################
# General config
#########################################################

# Name of the top-level package of your library.
#
# This is also the top level of its source tree, relative to the top-level project directory setup.py resides in.
#
libname = "pybalu"

# Choose build type.
#
build_type = "optimized"
# build_type="debug"

# Short description for package list on PyPI
#
SHORTDESC = "Python3 implementation of Computer Vision and Pattern Recognition library Balu"

# Long description for package homepage on PyPI
#
DESC = """Python3 implementation of Computer Vision and Pattern Recognition library Balu.

Made By Domingo Mery.
Implementation on Python3 done by Marco Bucchi.
"""

# Set up data files for packaging.
#
# Directories (relative to the top-level directory where setup.py resides) in which to look for data files.
datadirs = ("pybalu",)

# File extensions to be considered as data files. (Literal, no wildcards.)
dataexts = (".py",  ".pyx", ".pxd",  ".c", ".cpp", ".h",
            ".sh",  ".lyx", ".tex", ".txt", ".pdf")

# Standard documentation to detect (and package if it exists).
#
standard_docs = ["README", "LICENSE", "TODO", "CHANGELOG",
                 "AUTHORS"]  # just the basename without file extension
# commonly .md for GitHub projects, but other projects may use .rst or .txt (or even blank).
standard_doc_exts = [".md", ".rst", ".txt", ""]


#########################################################
# Init
#########################################################

# check for Python 3.6 or later
# http://stackoverflow.com/questions/19534896/enforcing-python-version-in-setup-py
if sys.version_info < (3, 6):
    sys.exit('Sorry, Python < 3.6 is not supported')

IS_WINDOWS = sys.platform.startswith("win")


USE_CYTHON = 'auto'             # True, 'auto' or False

if USE_CYTHON:
    try:
        from Cython.Build import cythonize
    except ImportError:
        if USE_CYTHON == 'auto':
            USE_CYTHON = False
        else:
            sys.exit(
                "Cython not found. Cython is needed to build the extension modules.")


#########################################################
# Definitions
#########################################################

# Define our base set of compiler and linker flags.
#
# This is geared toward x86_64, see
#    https://gcc.gnu.org/onlinedocs/gcc-4.6.4/gcc/i386-and-x86_002d64-Options.html
#
# Customize these as needed.
#
# Note that -O3 may sometimes cause mysterious problems, so we limit ourselves to -O2.

# Modules involving numerical computations
#
extra_compile_args_math_optimized = [
    '-march=native', '-O2', '-msse', '-msse2', '-mfma', '-mfpmath=sse']
extra_compile_args_math_debug = ['-march=native', '-O0', '-g']
extra_link_args_math_optimized = []
extra_link_args_math_debug = []

# Modules that do not involve numerical computations
#
extra_compile_args_nonmath_optimized = ['-O2']
extra_compile_args_nonmath_debug = ['-O0', '-g']
extra_link_args_nonmath_optimized = []
extra_link_args_nonmath_debug = []

# Additional flags to compile/link with OpenMP
#
openmp_compile_args = ['-fopenmp']
openmp_link_args = ['-fopenmp']


#########################################################
# Helpers
#########################################################

# Make absolute cimports work.
#
# See
#     https://github.com/cython/cython/wiki/PackageHierarchy
#
# For example: my_include_dirs = [np.get_include()]
my_include_dirs = ["."]
try:
    import numpy as np
    my_include_dirs.append(np.get_include())
except ModuleNotFoundError:
    pass


# Choose the base set of compiler and linker flags.
#
if build_type == 'optimized':
    my_extra_compile_args_math = extra_compile_args_math_optimized
    my_extra_compile_args_nonmath = extra_compile_args_nonmath_optimized
    my_extra_link_args_math = extra_link_args_math_optimized
    my_extra_link_args_nonmath = extra_link_args_nonmath_optimized
    my_debug = False
    print("build configuration selected: optimized")
elif build_type == 'debug':
    my_extra_compile_args_math = extra_compile_args_math_debug
    my_extra_compile_args_nonmath = extra_compile_args_nonmath_debug
    my_extra_link_args_math = extra_link_args_math_debug
    my_extra_link_args_nonmath = extra_link_args_nonmath_debug
    my_debug = True
    print("build configuration selected: debug")
else:
    raise ValueError(
        "Unknown build configuration '%s'; valid: 'optimized', 'debug'" % (build_type))


def declare_cython_extension(extName, use_math=False, use_openmp=False, include_dirs=None):
    """Declare a Cython extension module for setuptools.

Parameters:
    extName : str
        Absolute module name, e.g. use `mylibrary.mypackage.mymodule`
        for the Cython source file `mylibrary/mypackage/mymodule.pyx`.

    use_math : bool
        If True, set math flags and link with ``libm``.

    use_openmp : bool
        If True, compile and link with OpenMP.

Return value:
    Extension object
        that can be passed to ``setuptools.setup``.
"""
    extPath = extName.replace(".", os.path.sep)

    if USE_CYTHON:
        extPath += ".pyx"
    else:
        extPath += ".c"

    libraries = None

    if use_math:
        compile_args = list(my_extra_compile_args_math)  # copy
        link_args = list(my_extra_link_args_math)
        # link libm; this is a list of library names without the "lib" prefix
        if not IS_WINDOWS:
            libraries = ["m"]
    else:
        compile_args = list(my_extra_compile_args_nonmath)
        link_args = list(my_extra_link_args_nonmath)
        libraries = None  # value if no libraries, see setuptools.extension._Extension

    # OpenMP
    if use_openmp:
        compile_args.insert(0, openmp_compile_args)
        link_args.insert(0, openmp_link_args)

    # See
    #    http://docs.cython.org/src/tutorial/external.html
    #
    # on linking libraries to your Cython extensions.
    #
    return Extension(extName,
                     [extPath],
                     extra_compile_args=compile_args,
                     extra_link_args=link_args,
                     include_dirs=include_dirs,
                     libraries=libraries
                     )


# Gather user-defined data files
#
# http://stackoverflow.com/questions/13628979/setuptools-how-to-make-package-contain-extra-data-folder-and-all-folders-inside
#
datafiles = []


def getext(filename): return os.path.splitext(filename)[1]


for datadir in datadirs:
    datafiles.extend([(root, [os.path.join(root, f) for f in files if getext(f) in dataexts])
                      for root, dirs, files in os.walk(datadir)])


# Add standard documentation (README et al.), if any, to data files
#
detected_docs = []
for docname in standard_docs:
    for ext in standard_doc_exts:
        # relative to the directory in which setup.py resides
        filename = "".join((docname, ext))
        if os.path.isfile(filename):
            detected_docs.append(filename)
datafiles.append(('.', detected_docs))


# Extract __version__ from the package __init__.py
# (since it's not a good idea to actually run __init__.py during the build process).
#
# http://stackoverflow.com/questions/2058802/how-can-i-get-the-version-defined-in-setup-py-setuptools-in-my-package
#
init_py_path = os.path.join(libname, '__init__.py')
version = 'ERROR'
try:
    with open(init_py_path) as f:
        for line in f:
            if line.startswith('__version__'):
                version = ast.parse(line).body[0].value.s
                break
        else:
            print("WARNING: Version information not found in '%s', using placeholder '%s'" % (
                init_py_path, version), file=sys.stderr)
except FileNotFoundError:
    print("WARNING: Could not find file '%s', using placeholder version information '%s'" % (
        init_py_path, version), file=sys.stderr)

#########################################################
# Set up modules
#########################################################

# declare Cython extension modules here
#
# ext_module_feature_analysis = declare_cython_extension( "pybalu.feature_analysis", use_math=True, include_dirs=my_include_dirs)
ext_module_fex_flusser = declare_cython_extension(
    "pybalu.feature_extraction.flusser_utils", use_math=True, include_dirs=my_include_dirs)
ext_module_fex_geometric = declare_cython_extension(
    "pybalu.feature_extraction.geometric_utils", use_math=True, include_dirs=my_include_dirs)
ext_module_fex_haralick = declare_cython_extension(
    "pybalu.feature_extraction.haralick_utils", use_math=True, include_dirs=my_include_dirs)
ext_module_fex_hog = declare_cython_extension(
    "pybalu.feature_extraction.hog_utils", use_math=True, include_dirs=my_include_dirs)

# this is mainly to allow a manual logical ordering of the declared modules
#
cython_ext_modules = [ext_module_fex_flusser,
                      ext_module_fex_geometric,
                      ext_module_fex_haralick,
                      ext_module_fex_hog]

# Call cythonize() explicitly, as recommended in the Cython documentation. See
#     http://cython.readthedocs.io/en/latest/src/reference/compilation.html#compiling-with-distutils
#
# This will favor Cython's own handling of '.pyx' sources over that provided by setuptools.
#
# Note that my_ext_modules is just a list of Extension objects. We could add any C sources (not coming from Cython modules) here if needed.
# cythonize() just performs the Cython-level processing, and returns a list of Extension objects.
#
if USE_CYTHON:
    my_ext_modules = cythonize(
        cython_ext_modules, include_path=my_include_dirs, gdb_debug=my_debug)
else:
    my_ext_modules = cython_ext_modules


#########################################################
# Call setup()
#########################################################

setup(
    name="pybalu",
    version=version,
    author="Marco Bucchi",
    author_email="mabucchi@uc.cl",
    url="https://github.com/mbucchi",

    description=SHORTDESC,
    long_description=DESC,

    # CHANGE THIS
    license="MIT",

    # free-form text field; http://stackoverflow.com/questions/34994130/what-platforms-argument-to-setup-in-setup-py-does
    platforms=["Linux"],

    # See
    #    https://pypi.python.org/pypi?%3Aaction=list_classifiers
    #
    # for the standard classifiers.
    #
    # Remember to configure these appropriately for your project, especially license!
    #
    classifiers=["Development Status :: 3 - Alpha",
                 "Environment :: Console",
                 "Intended Audience :: Developers",
                 "Intended Audience :: Science/Research",
                 # not a standard classifier; CHANGE THIS
                 "License :: OSI Approved :: MIT License",
                 "Operating System :: POSIX :: Linux",
                 "Programming Language :: Cython",
                 "Programming Language :: Python",
                 "Programming Language :: Python :: 3",
                 "Programming Language :: Python :: 3.6",
                 "Topic :: Scientific/Engineering",
                 "Topic :: Scientific/Engineering :: Mathematics",
                 ],

    # See
    #    http://setuptools.readthedocs.io/en/latest/setuptools.html
    #
    install_requires=[
        "numpy>=1.16.1",
        "scipy>=1.1.0",
        "imageio>=2.5.0",
        "Pillow>=7.2.0",
        "scikit-image>=0.17.2",
        "scikit-learn>=0.22.2",
        "tqdm>=4.29.1"
    ],
    setup_requires=[
        "cython>=0.29.6",
        "numpy>=1.16.1",
        "scipy>=1.1.0",
        "imageio>=2.5.0",
        "Pillow>=7.2.0",
        "scikit-image>=0.17.2",
        "scikit-learn>=0.22.2",
        "tqdm>=4.29.1"
    ],
    python_requires='!=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*, <4',
    provides=["pybalu"],

    # keywords for PyPI (in case you upload your project)
    #
    # e.g. the keywords your project uses as topics on GitHub, minus "python" (if there)
    #
    keywords=["image pattern recognition computer vision"],

    # All extension modules (list of Extension objects)
    #
    ext_modules=my_ext_modules,

    # Declare packages so that  python -m setup build  will copy .py files (especially __init__.py).
    #
    # This **does not** automatically recurse into subpackages, so they must also be declared.
    #
    packages=["pybalu",
              "pybalu.classification",
              "pybalu.data_selection",
              "pybalu.feature_analysis",
              "pybalu.feature_extraction",
              "pybalu.feature_selection",
              "pybalu.feature_transformation",
              "pybalu.img_processing",
              "pybalu.io",
              "pybalu.misc",
              "pybalu.performance_eval"],

    # Install also Cython headers so that other Cython modules can cimport ours
    #
    # Fileglobs relative to each package, **does not** automatically recurse into subpackages.
    #
    # FIXME: force sdist, but sdist only, to keep the .c files (this puts them also in the bdist)
    package_data={'pybalu': ['*.c'],
                  'pybalu.classification': ['*.c'],
                  'pybalu.data_selection': ['*.c'],
                  'pybalu.feature_analysis': ['*.c'],
                  'pybalu.feature_extraction': ['*.c'],
                  'pybalu.feature_selection': ['*.c'],
                  'pybalu.feature_transformation': ['*.c'],
                  'pybalu.img_processing': ['*.c'],
                  'pybalu.io': ['*.c'],
                  'pybalu.misc': ['*.c'],
                  'pybalu.performance_eval': ['*.c'],
                  'pybalu.utils': ['*.c']},

    # Disable zip_safe, because:
    #   - Cython won't find .pxd files inside installed .egg, hard to compile libs depending on this one
    #   - dynamic loader may need to have the library unzipped to a temporary directory anyway (at import time)
    #
    zip_safe=False,

    # Custom data files not inside a Python package
    data_files=datafiles
)



__all__ = ['hugeo_features', 'HuGeoExtractor']

import numpy as np

from pybalu.base import FeatureExtractor
# pylint: disable=no-name-in-module
from .geometric_utils import moments
# pylint: enable=no-name-in-module


hu_labels = ['Hu-moment 1',
             'Hu-moment 2',
             'Hu-moment 3',
             'Hu-moment 4',
             'Hu-moment 5',
             'Hu-moment 6',
             'Hu-moment 7']


def hugeo_features(image, *, show=False, labels=False):
    '''\
hugeo_features(image, *, show=False, labels=False)

Return an array of with the 7 Hu moments of a binary image.

Parameters
----------
image : a numpy 2 dimensional int array
    It represents a binary image. Non binary arrays will return nonsensical results.
show : bool, optional
    Wether to print or not messages during execution
labels : bool, optional
    Wether to return a second array that contains the label of each value. 

Returns
-------
labels: ndarray, optional
    A one dimensional string ndarray that contains the labels to each of the features.
    This array is only returned if `labels` is set to True.
features: ndarray
    A float ndarray that contains the 7 Hu moments extracted from `image`.

See Also
--------
basic_geo : Extract the standard geometric features from a binary image.
fit_ellipse : <Not Implemented Yet>
flusser : Extract the four Flusser moments from binary image.
gupta : Extract the three Gupta moments from a binary image.

Examples
--------
Load an image and get its binary representation, then proceed to get its features:

>>> from pybalu.feature_extraction import hugeo_features
>>> from pybalu.img_processing import segbalu
>>> from pybalu.io import imread
>>> img = imread('testimg.png')
>>> binary_img, _, _ = segbalu(img)
>>> features = hugeo_features(binary_img)

Print a binary image features:

>>> from pybalu.io import print_features
>>> labels, features = hugeo_features(binary_img, labels=True)
>>> print_features(labels, features)
Hu-moment 1:  0.38161
Hu-moment 2:  0.00845
Hu-moment 3:  0.00560
Hu-moment 4:  0.00057
Hu-moment 5: -0.00000
Hu-moment 6: -0.00005
Hu-moment 7: -0.00000
'''
    if show:
        print('--- extracting Hu moments...')

    m = moments(image, centered=True)

    area = m[0, 0]
    area_sq = area ** 2
    area_25 = area ** 2.5

    n02 = m[0, 2] / area_sq
    n20 = m[2, 0] / area_sq
    n11 = m[1, 1] / area_sq
    n12 = m[1, 2] / area_25
    n21 = m[2, 1] / area_25
    n03 = m[0, 3] / area_25
    n30 = m[3, 0] / area_25

    f1 = n20 + n02
    f2 = (n20-n02)**2 + 4*n11**2
    f3 = (n30-3*n12)**2+(3*n21-n03)**2
    f4 = (n30+n12)**2+(n21+n03)**2
    f5 = (n30-3*n12)*(n30+n12)*((n30+n12)**2 - 3*(n21+n03)**2) + \
        (3*n21-n03)*(n21+n03)*(3*(n30+n12)**2 - (n21+n03)**2)
    f6 = (n20-n02)*((n30+n12)**2 - (n21+n03)**2) + 4*n11*(n30+n12)*(n21+n03)
    f7 = (3*n21-n03)*(n30+n12)*((n30+n12)**2 - 3*(n21+n03)**2) - \
        (n30-3*n12)*(n21+n03)*(3*(n30+n12)**2 - (n21+n03)**2)

    hu_moments = np.array([f1, f2, f3, f4, f5, f6, f7])

    if labels:
        return np.array(hu_labels), hu_moments
    return hu_moments


class HuGeoExtractor(FeatureExtractor):

    def transform(self, X):
        return np.array([hugeo_features(x) for x in self._get_iterator(X)])

    def get_labels(self):
        return np.array(hu_labels)



__all__ = ['flusser_features', 'FlusserExtractor']
import numpy as np
from .flusser_utils import flusser as _flusser, f_labels
from pybalu.base import FeatureExtractor


def flusser_features(image, show=False, labels=False):
    '''\
flusser_features(image, *, show=False, labels=False)

Return an array of with the the four Flusser moments extracted from a binary image.

Parameters
----------
image : a numpy 2 dimensional int array
    It represents a binary image. Non binary arrays will return nonsensical results.
show : bool, optional
    Wether to print or not messages during execution
labels : bool, optional
    Wether to return a second array that contains the label of each value. 

Returns
-------
labels: ndarray, optional
    A one dimensional string ndarray that contains the labels to each of the features.
    This array is only returned if `labels` is set to True.
features: ndarray
    A float ndarray that contains all 4 Flusser moments extracted from the image.

See Also
--------
fit_ellipse : <Not Implemented Yet>
gupta : Extract the three Gupta moments from binary image.
hugeo : Extract the seven Hu moments from a binary image.
standard : <Not Implemented Yet>

Examples
--------
Load an image and get its binary representation, then proceed to get its features:

>>> from pybalu.feature_extraction import flusser_features
>>> from pybalu.img_processing import segbalu
>>> from pybalu.io import imread
>>> img = imread('testimg.png')
>>> binary_img, _, _ = segbalu(img)
>>> features = flusser_features(binary_img)

Print a binary image features:

>>> from pybalu.io import print_features
>>> labels, features = flusser_features(binary_img, labels=True)
>>> print_features(labels, features)
Flusser-moment 1:  0.03429
Flusser-moment 2: -0.00000
Flusser-moment 3: -0.00009
Flusser-moment 4:  0.00003
'''
    return _flusser(image, show=show, labels=labels)


class FlusserExtractor(FeatureExtractor):

    def transform(self, X):
        return np.array([flusser_features(x) for x in self._get_iterator(X)])

    def get_labels(self):
        return np.array(f_labels)



__all__ = ['fourier_des_features', 'FourierDesExtractor']

import numpy as np
from pybalu.base import FeatureExtractor
# pylint: disable=no-name-in-module
from .geometric_utils import bw_boundaries
# pylint: enable=no-name-in-module


def fourier_des_features(image, *, n_des=16, show=False, labels=False):
    '''\
fourier_des_features(image, *, n_des=16, show=False, labels=False)

Return an array of with the Fourier descriptors of a binary image.

Parameters
----------
image : a numpy 2 dimensional int array
    It represents a binary image. Non binary arrays will return nonsensical results.
n_des : int, optional
    The number of descriptors to extract from the image. Default is 16
show : bool, optional
    Wether to print or not messages during execution
labels : bool, optional
    Wether to return a second array that contains the label of each value. 

Returns
-------
labels: ndarray, optional
    A one dimensional string ndarray that contains the labels to each of the features.
    This array is only returned if `labels` is set to True.
features: ndarray
    A float ndarray that contains `n_des` Fourier descriptors extracted from `image`.

See Also
--------
fit_ellipse : <Not Implemented Yet>
flusser : Extract the four Flusser moments from binary image.
gupta : Extract the three Gupta moments from binary image.
hugeo : Extract the seven Hu moments from a binary image.

Examples
--------
Load an image and get its binary representation, then proceed to get its features:

>>> from pybalu.feature_extraction import fourier_des_features
>>> from pybalu.img_processing import segbalu
>>> from pybalu.io import imread
>>> img = imread('testimg.png')
>>> binary_img, _, _ = segbalu(img)
>>> features = fourier_des_features(binary_img)

Print a binary image features:

>>> from pybalu.io import print_features
>>> labels, features = fourier_des_features(binary_img, labels=True)
>>> print_features(labels, features)
Fourier-des  1:  0.66428
Fourier-des  2:  0.92453
Fourier-des  3:  1.00341
Fourier-des  4:  0.49639
Fourier-des  5:  1.05206
Fourier-des  6:  0.34365
Fourier-des  7:  0.06147
Fourier-des  8:  0.12616
Fourier-des  9:  0.01778
Fourier-des 10:  0.10414
Fourier-des 11:  0.04698
Fourier-des 12:  0.13555
Fourier-des 13:  0.42163
Fourier-des 14:  0.06778
Fourier-des 15:  0.01811
Fourier-des 16:  0.05512
'''
    if show:
        print('--- extracting Fourier descriptors...')

    B = bw_boundaries(image)
    V = B[:, 1] + 1j * B[:, 0]
    m = B.shape[0]

    r = np.zeros(m, dtype=complex)
    phi = np.zeros(m)
    dphi = np.zeros(m)
    l = np.zeros(m)
    dl = np.zeros(m)

    r[0] = V[0] - V[m-1]
    r[1:] = V[1:] - V[:m-1]

    dl = np.abs(r)
    phi = np.angle(r)

    dphi[:m-1] = np.mod(phi[1:] - phi[:m-1] + np.pi, 2 * np.pi) - np.pi
    dphi[m-1] = np.mod(phi[0] - phi[m-1] + np.pi, 2 * np.pi) - np.pi

    l[0] = dl[0]
    for k in range(1, m):
        l[k] = l[k-1] + dl[k]

    l = l * (2 * np.pi / l[m-1])
    descriptors = np.zeros(n_des)

    for n in range(1, n_des + 1):
        an = (dphi * np.sin(l * n)).sum()
        bn = (dphi * np.cos(l * n)).sum()
        an = -an / n / np.pi
        bn = bn / n / np.pi
        imagi = an + 1j * bn
        descriptors[n-1] = np.abs(imagi)

    if labels:
        return np.array([f'Fourier-des {n+1:>2d}' for n in range(n_des)]), descriptors
    return descriptors


class FourierDesExtractor(FeatureExtractor):
    def __init__(self, *, n_des=16):
        self.n_des = n_des

    def transform(self, X):
        params = self.get_params()
        return np.array([fourier_des_features(x, **params) for x in self._get_iterator(X)])

    def get_labels(self):
        return np.array([f'Fourier-des {n+1:>2d}' for n in range(self.n_des)])



__all__ = ['ImageSet']

import os
import re
import collections
import numpy as np
from .imread import imread


class ImageSet:
    def __init__(self, path, *, extension=None, prefix=None, start=None, stop=None,
                 flatten=True,  imloader=None, imloader_opts=None, filenames=None):

        if isinstance(path, ImageSet):
            other = path
            self._directory = other._directory
            self._file_re = re.compile(other._file_re.pattern)
            self._imloader_opts = dict(other._imloader_opts)
            self._flatten = other._flatten
            self._imloader = other._imloader
            if filenames is not None:
                self._filenames = list(filenames)
                self._slice = slice(len(filenames))
                return

            if start is not None or stop is not None:
                other_start = other._slice.start or 0
                _start = other_start + start if start is not None else other_start
                _stop = other._slice.stop if stop is None else other_start + stop
                self._slice = slice(_start, min(_stop, other._slice.stop))
            else:
                self._slice = other._slice

            if self._flatten:
                self._dir_filenames = None
                self._filenames = other._filenames[start: stop]
            else:
                self._dir_filenames = other._dir_filenames[start: stop]
                self._filenames = list(np.hstack(self._dir_filenames))

            # TODO: should we warn if the resulting set is empty?
            return

        if extension is None:
            extension = '.*?'
        else:
            extension = extension.split('.')[-1]

        if prefix is None:
            prefix = ''

        self._directory = path
        self._file_re = re.compile(fr'{prefix}.*\.{extension}$')
        self._imloader_opts = imloader_opts or dict()
        self._imloader = imloader or imread
        self._flatten = flatten
        if flatten:
            self._dir_filenames = None
            self._filenames = [os.path.join(dirname, filename)
                               for dirname, _, files in os.walk(path)
                               for filename in files
                               if self._file_re.match(filename)][start: stop]
            self._slice = slice(start or 0, len(self._filenames))
        else:
            self._dir_filenames = [[os.path.join(dirname, filename)
                                    for filename in files if self._file_re.match(filename)]
                                   for dirname, _, files in os.walk(path)][start: stop]
            self._dir_filenames = [
                dirs for dirs in self._dir_filenames if len(dirs) > 0]
            self._filenames = list(np.hstack(self._dir_filenames))
            self._slice = slice(start or 0, len(self._dir_filenames))

    def set_loader(self, imloader):
        self._imloader = imloader

    def is_flat(self):
        return self._flatten

    def __iter__(self):
        return ((filename, self._imloader(filename, **self._imloader_opts)) for filename in self._filenames)

    def get_filenames(self):
        return list(self._filenames)

    def get_images(self):
        return list(self.iter_images())

    def iter_images(self):
        if self._flatten:
            return (self._imloader(filename, **self._imloader_opts) for filename in self._filenames)
        else:
            return (np.array([self._imloader(filename, **self._imloader_opts) for filename in _dir]) for _dir in self._dir_filenames)

    def __len__(self):
        if self._flatten:
            return len(self._filenames)
        else:
            return len(self._dir_filenames)

    @property
    def shape(self):
        if self._flatten:
            return (len(self._filenames),)
        else:
            return len(self._dir_filenames), len(self._filenames)

    def __getitem__(self, obj):
        if isinstance(obj, int):
            idx = int(obj)
            if self._flatten:
                return self._imloader(self._filenames[idx], **self._imloader_opts)
            else:
                return np.array([self._imloader(filename, **self._imloader_opts) for filename in self._dir_filenames[idx]])
        elif isinstance(obj, slice):
            _slice = obj
            return ImageSet(self, start=_slice.start, stop=_slice.stop)
        elif isinstance(obj, collections.Iterable):
            filenames = np.array(self._filenames)[obj]
            return ImageSet(self, filenames=filenames)
        else:
            raise TypeError(f"Expected a slice or an int, not {type(obj)}")

    def __repr__(self):
        attrs = [f"path='{self._directory}'",
                 f"pattern='{self._file_re.pattern}'"]
        if self._slice.start is not None:
            attrs.append(f'start={self._slice.start}')
        if self._slice.stop is not None:
            attrs.append(f'stop={self._slice.stop}')
        return f'{self.__class__.__name__}({", ".join(iter(attrs))})'



__all__ = ['imread']


import numpy as _np
from imageio import imread as _imread


def imread(filename, *, normalize=False, flatten=False):
    '''\
    imread(filename, flatten=False)

    Loads an image as a numerical matrix of either 2 or 3 dimensions.

    Parameters
    ----------
    filename: string
        The path of the image to load
    normalize: boolean, optional
        If set to true, the return value will be an array with float values between 0 and 1.
        If set to false, the reurn value will be an array with uint8 values between 0 and 255.
        default value is False.
    flatten: boolean, optional
        If set to true, the return value is a 2 dimensional ndarray with the grayscale
        representation of the loaded image.

    Returns
    -------
    image: ndarray
        matrix representation of the loaded image, either 2 or 3 dimensional depending on the
        file and the `flatten` parameter.

    Examples
    --------
    (TODO)
    '''
    img = _imread(filename)
    if flatten:
        img = img @ [0.299, 0.587, 0.114]
    if normalize:
        return (img / 255)
    return img.astype(_np.uint8)



__all__ = ["im2col", "im2row"]

import numpy as _np


def im2col(image, n, m):
    '''\
    im2col(image, n, m)

    Rearranges discrete image blocks of size m-by-n into columns, and returns the concatenated columns. 
    im2col pads `image` with zeros, if necessary.

    Parameters 
    ----------
    image: a numpy 2 dimensional array
        Represents the input image of dimensions N-by-M
    n: non-zero positive integer value
        The size of the blocks on the first dimension.
    m: non-zero positive integer value
        The size of the blocks on the second dimension.

    Returns
    -------
    blocks: 2 dimensional ndarray of size (n * m)-by-(ceil(N / n) * ceil(M / m)) of the same type as `image`.
        Each column represents a diferent block of (n * m) values on the original image.
        The blocks are iterated from (0, 0) increasing initially in the first dimension.
        ie: (0, 0), (n, 0), (2*n, 0), ... , (x*n, 0), (0, m), (n, m), ... ,(x*n, m), ...

    Examples
    --------
    Generate a 4-by-5 matrix and transform it to columns in 2-by-3 blocks

    >>> import numpy as np
    >>> from pybalu.misc import im2col
    >>> img = np.arange(20).reshape(4, 5)
    >>> print(img)
    [[ 0  1  2  3  4]
     [ 5  6  7  8  9]
     [10 11 12 13 14]
     [15 16 17 18 19]]
    >>> print(im2col(img, 2, 3))
    [[ 0 10  3 13]
     [ 5 15  8 18]
     [ 1 11  4 14]
     [ 6 16  9 19]
     [ 2 12  0  0]
     [ 7 17  0  0]]
    '''
    image = image.T
    n, m = m, n
    N, M = image.shape
    dn, rn = divmod(N, n)
    dm, rm = divmod(M, m)
    if rn or rm:
        if rn:
            _N = (dn+1)*n
        else:
            _N = N
        if rm:
            _M = (dm+1)*m
        else:
            _M = M
        _img = _np.zeros((_N, _M), dtype=image.dtype)
        _img[:N, :M] = image
    else:
        _img = _np.ascontiguousarray(image)

    arr_shape = _np.array(_img.shape)
    block_shape = _np.array([n, m])
    new_shape = tuple(arr_shape // block_shape) + tuple(block_shape)
    new_strides = tuple(_img.strides * block_shape) + _img.strides
    blocks = _np.lib.stride_tricks.as_strided(
        _img, shape=new_shape, strides=new_strides)
    _n, _m, *_ = new_shape
    return blocks.reshape(_n*_m, -1).T


def im2row(image, n, m):
    '''\
    im2row(image, n, m)

    Divides the given image into overlaping blocks that slide accross the image and returns the 
    concatenated rows. The block size is such that it is the biggest possible sliding window that
    divides the image into n-by-m distinct blocks. Depending on dimensions, it may or may not
    cover the whole image.

    Parameters 
    ----------
    image: a numpy 2 dimensional array
        Represents the input image of dimensions N-by-M
    n: non-zero positive integer value
        The number of sliding blocks in the first dimension.
    m: non-zero positive integer value
        The size of sliding blocks in the second dimension

    Returns
    -------
    windows: 2 dimensional ndarray of the same type as `image`.
        of size (n * m)-by-(N // (n + 1) * M // (m + 1) * 4). Each row represents a 
        diferent view of the sliding window over the original image. The window is  
        moved from (0, 0) increasing initially in the first dimension.
        ie: (0, 0), (N // (n + 1), 0), ... , (n * N // (n + 1), 0), 
        (0, M // (m + 1)), (, ... ,((n * N // (n + 1),  M // (m + 1)), ...

    Examples
    --------
    Generate a 4-by-4 matrix and transform it to rows in a 3-by-3 window division

    >>> import numpy as np
    >>> from pybalu.misc import im2row
    >>> img = np.arange(16).reshape(4, 4)
    >>> print(img)
    [[ 0  1  2  3]
     [ 4  5  6  7]
     [ 8  9 10 11]
     [12 13 14 15]]
    >>> print(im2row(img, 3, 3))
    [[ 0  1  4  5]
     [ 1  2  5  6]
     [ 2  3  6  7]
     [ 4  5  8  9]
     [ 5  6  9 10]
     [ 6  7 10 11]
     [ 8  9 12 13]
     [ 9 10 13 14]
     [10 11 14 15]]
    '''
    _img = _np.ascontiguousarray(image)
    N, M = _img.shape
    hh = N // (n + 1)
    hw = M // (m + 1)
    vs, hs = _img.strides

    new_shape = (n, m, hh*2, hw*2)
    new_strides = (vs * hh, hs * hw, vs, hs)
    windows = _np.lib.stride_tricks.as_strided(
        _img, shape=new_shape, strides=new_strides)
    return windows.reshape(n * m, -1)



__all__ = ['lbp_features', 'LBPExtractor']

import numpy as np
from pybalu.misc import im2col
from skimage.feature import local_binary_pattern as _lbp

from pybalu.base import FeatureExtractor


def lbp_features(image, region=None, *, show=False, labels=False, **kwargs):
    '''\
    lbp_features(image, region=None, *, show=False, labels=False, **kwargs)

    Calculates the Local Binary Patterns over a regular grid of patches. It returns an array
    of uniform lbp82 descriptors for `image`, made by the concatenating histograms of each grid 
    cell in the image. Grid size is `hdiv` * `vdiv`

    Parameters 
    ----------
    image: 2 dimensional ndarray
        It represents a grayscale image or just one dimension of color (eg: green channel)
    region: 2 dimensional ndarray, optional
        If not None, must be of same dimensions an image and of a bool-like type. All the pixels
        not set in this region will be set to 0 on `image` before performing LBP calculations.
    hdiv: positive integer
        Number of horizontal divisions to perform on image.
    vdiv: positive integer
        Number of vertical divisions to perform on image.
    samples: positive integer, optional
        Number of circularly symmetric neighbour set points (quantization of the angular space).
        default value is 8 (all neighbours in 2d)
    norm: bool, optional
        If set to True, the output array is normalized so that the sum of all its features 
        equals 1. Default value is False.
    mapping: string, optional
        Reprsents the kind of LBP performed over each block. Options are:
            - 'default': original local binary pattern which is gray scale but not
               rotation invariant.
            - 'ror': extension of default implementation which is gray scale and
               rotation invariant.
            - 'uniform': improved rotation invariance with uniform patterns and
               finer quantization of the angular space which is gray scale and
               rotation invariant.
            - 'nri_uniform': non rotation-invariant uniform patterns variant
               which is only gray scale invariant.
        default value is 'default'.
    radius: integer, optional
        Radius of circle (spatial resolution of the operator). Default is set depending on number 
        of samples
    ret_centers: bool, optional
        If set to True, an array with the centers of each block over which LBP was performed is 
        returned. Default value is False.
    show: bool, optional
        Wether to print or not messages during execution. Default is False.
    labels: bool, optional
        Wether to return a second array that contains the label of each value. 

    Returns
    -------
    labels: ndarray, optional
        A one dimensional string ndarray that contains the labels to each of the features.
        This array is only returned if `labels` is set to True.
    features: ndarray
        A numeric ndarray that contains (`hdiv`*`vdiv`) * `num_patterns` features extracted from 
        `image`. `num_patterns` depends on `mapping` and `samples` and is usually 10, 59 or 256.
    x_centers: integer ndarray, optional
        A one dimensional array of size `hdiv`*`vdiv` that contains the x center 
        coordinate of the blocks generated on image division. only returned if `ret_centers` is 
        set to True.
    y_centers: integer ndarray, optional
        A one dimensional array of size `hdiv`*`vdiv` that contains the y center 
        coordinate of the blocks generated on image division. only returned if `ret_centers` is 
        set to True.

    Examples
    --------
    ( TODO )
    '''
    vdiv = kwargs.pop('vdiv', None)
    hdiv = kwargs.pop('hdiv', None)
    if vdiv is None or hdiv is None:
        raise ValueError('`vdiv` and `hdiv` must be given to lbp.')

    if region is None:
        region = np.ones_like(image)

    samples = kwargs.pop('samples', 8)
    normalize = kwargs.pop('norm', False)
    integral = kwargs.pop('integral', False)
    max_d = kwargs.pop('max_d', None)
    if integral and max_d is None:
        raise ValueError('`max_d` must be set if `integral` is set to True.')

    weight = kwargs.pop('weight', 0)
    mapping = kwargs.pop('mapping', 'default')

    if mapping == 'ror' or mapping == 'default':
        num_patterns = 2 ** samples
    elif mapping == 'uniform':
        num_patterns = samples + 2
    elif mapping == 'nri_uniform':
        num_patterns = 59
    else:
        raise ValueError(f"Unknown mapping: '{mapping}'")

    radius = kwargs.pop('radius', None)
    if radius is None:
        radius = np.log(samples) / np.log(2) - 1

    ret_centers = kwargs.pop('ret_centers', False)

    if len(kwargs) > 0:
        unknowns = "'" + "', '".join(kwargs.keys()) + "'"
        raise ValueError(f"Unknown options given to lbp: {unknowns}")

    if show:
        print('--- extracting local binary patterns features...')
    label = 'LBP'

    # set pixels not within region to 0
    image = image.copy()
    image[~region.astype(bool)] = 0

    code_img = _lbp(image, P=samples, R=radius, method=mapping)
    n, m = code_img.shape
    N, M = image.shape
    Ilbp = np.zeros_like(image)
    i1 = round((N - n)/2)
    j1 = round((M - m)/2)
    Ilbp[i1:i1+n, j1:j1+m] = code_img

# TODO:
#     if integral:
#         hx = inthist(Ilbp+1, max_d)

    ylen = int(np.ceil(n / vdiv))
    xlen = int(np.ceil(m / hdiv))
    grid_img = im2col(code_img, ylen, xlen) + 1

    if weight > 0:
        label = 'W-' + label
        pass
    else:
        desc = np.vstack([np.histogram(grid_img[:, i], num_patterns)[0]
                          for i in range(grid_img.shape[1])])

    lbp_feats = desc.ravel()
    N, M = desc.shape

    if normalize:
        lbp_feats = lbp_feats / lbp_feats.sum()

    if not labels and not ret_centers:
        return lbp_feats

    if labels:
        lbp_labels = []
        for i in range(N):
            for j in range(M):
                lbp_labels.append(
                    f"{label}({i+1},{j+1:>2d}) [{samples},'{mapping}']")
        ret = np.array(lbp_labels), lbp_feats
    else:
        ret = (lbp_feats,)

    if ret_centers:
        dx = 1 / hdiv
        dy = 1 / vdiv
        x = np.linspace(dx / 2, 1 - dx / 2, hdiv)
        y = np.linspace(dy / 2, 1 - dy / 2, vdiv)
        ret = (ret,) + (x, y)

    return ret


class LBPExtractor(FeatureExtractor):
    def __init__(self, *, hdiv=None, vdiv=None, samples=8, norm=False, mapping="default", radius=None):
        self.hdiv = hdiv
        self.vdiv = vdiv
        self.samples = samples
        self.norm = norm
        self.mapping = mapping
        self.radius = radius

    def transform(self, X):
        params = self.get_params()
        return np.array([lbp_features(x, **params) for x in self._get_iterator(X)])

    def get_labels(self):
        return "LBP"
